# Copyright (c) Acconeer AB, 2023
# All rights reserved

from .dialog import SaveDialogWithPreview
from .preview import PresentationType, PresentationWindow, PresenterFunc
from .set_config_presenter import set_config_presenter



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

from typing import List, Optional, Sequence

import h5py
import numpy as np

from ._ref_app import RefAppResult


_ALL_REF_APP_RESULT_FIELDS = (
    "zone_limits",
    "presence_detected",
    "max_presence_zone",
    "total_zone_detections",
    "inter_presence_score",
    "inter_zone_detections",
    "max_inter_zone",
    "intra_presence_score",
    "intra_zone_detections",
    "max_intra_zone",
)


class PhonyNoneSeries:
    def __next__(self) -> None:
        return None

    def __iter__(self) -> PhonyNoneSeries:
        return self


def _serialize_optional_uint(num_list: List[Optional[int]]) -> List[int]:
    return [num if num is not None else -1 for num in num_list]


def _deserialize_optional_uint(num: int) -> Optional[int]:
    return None if num == -1 else num


class RefAppResultListH5Serializer:
    """
    Reads or writes a smart presence RefAppResult from/to a given h5py.Group
    """

    def __init__(
        self,
        group: h5py.Group,
        fields: Sequence[str] = _ALL_REF_APP_RESULT_FIELDS,
        allow_missing_fields: bool = False,
    ) -> None:
        """
        :param group: The H5 group that will have datasets read/written to it
        :param fields: The fields to serialize. Has no effect on deserialization
        :param allow_missing_fields:
            Whether it's acceptable to break type safety during de-serialization
            by replacing missing fields by ``None``. Has no effect on serialization.
        """
        self.group = group
        self.fields = fields
        self.allow_missing_fields = allow_missing_fields

    def serialize(self, results: List[RefAppResult]) -> None:
        if "zone_limits" in self.fields:
            self.group.create_dataset(
                "zone_limits",
                dtype=float,
                data=np.array([res.zone_limits for res in results]),
                track_times=False,
            )

        if "presence_detected" in self.fields:
            self.group.create_dataset(
                "presence_detected",
                dtype=bool,
                data=np.array([res.presence_detected for res in results]),
                track_times=False,
            )

        if "max_presence_zone" in self.fields:
            data = np.array(_serialize_optional_uint([res.max_presence_zone for res in results]))

            self.group.create_dataset(
                "max_presence_zone",
                dtype=int,
                data=data,
                track_times=False,
            )

        if "total_zone_detections" in self.fields:
            self.group.create_dataset(
                "total_zone_detections",
                dtype=int,
                data=np.array([res.total_zone_detections for res in results]),
                track_times=False,
            )

        if "inter_presence_score" in self.fields:
            self.group.create_dataset(
                "inter_presence_score",
                dtype=float,
                data=np.array([res.inter_presence_score for res in results]),
                track_times=False,
            )

        if "inter_zone_detections" in self.fields:
            self.group.create_dataset(
                "inter_zone_detections",
                dtype=int,
                data=np.array([res.inter_zone_detections for res in results]),
                track_times=False,
            )

        if "max_inter_zone" in self.fields:
            data = np.array(_serialize_optional_uint([res.max_inter_zone for res in results]))

            self.group.create_dataset(
                "max_inter_zone",
                dtype=int,
                data=data,
                track_times=False,
            )

        if "intra_presence_score" in self.fields:
            self.group.create_dataset(
                "intra_presence_score",
                dtype=float,
                data=np.array([res.intra_presence_score for res in results]),
                track_times=False,
            )

        if "intra_zone_detections" in self.fields:
            self.group.create_dataset(
                "intra_zone_detections",
                dtype=int,
                data=np.array([res.intra_zone_detections for res in results]),
                track_times=False,
            )

        if "max_intra_zone" in self.fields:
            data = np.array(_serialize_optional_uint([res.max_intra_zone for res in results]))

            self.group.create_dataset(
                "max_intra_zone",
                dtype=int,
                data=data,
                track_times=False,
            )

    def deserialize(self, _: None) -> List[RefAppResult]:
        groups = (
            self.group.get("zone_limits", PhonyNoneSeries()),
            self.group.get("presence_detected", PhonyNoneSeries()),
            self.group.get("max_presence_zone", PhonyNoneSeries()),
            self.group.get("total_zone_detections", PhonyNoneSeries()),
            self.group.get("inter_presence_score", PhonyNoneSeries()),
            self.group.get("inter_zone_detections", PhonyNoneSeries()),
            self.group.get("max_inter_zone", PhonyNoneSeries()),
            self.group.get("intra_presence_score", PhonyNoneSeries()),
            self.group.get("intra_zone_detections", PhonyNoneSeries()),
            self.group.get("max_intra_zone", PhonyNoneSeries()),
        )

        if any(isinstance(g, PhonyNoneSeries) for g in groups) and not self.allow_missing_fields:
            raise ValueError("Some fields are missing while 'allow_missing_fields' is False.")

        if all(isinstance(group, PhonyNoneSeries) for group in groups):
            return []

        return [
            RefAppResult(
                zone_limits=zone_limits,
                presence_detected=presence_detected,
                max_presence_zone=_deserialize_optional_uint(max_presence_zone),
                total_zone_detections=total_zone_detections,
                inter_presence_score=inter_presence_score,
                inter_zone_detections=inter_zone_detections,
                max_inter_zone=_deserialize_optional_uint(max_inter_zone),
                intra_presence_score=intra_presence_score,
                intra_zone_detections=intra_zone_detections,
                max_intra_zone=_deserialize_optional_uint(max_intra_zone),
            )
            for (
                zone_limits,
                presence_detected,
                max_presence_zone,
                total_zone_detections,
                inter_presence_score,
                inter_zone_detections,
                max_inter_zone,
                intra_presence_score,
                intra_zone_detections,
                max_intra_zone,
            ) in zip(*groups)
        ]



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from . import icons, plugin_components, utils
from .app_model_viewer import AppModelViewer
from .main_window import MainWindow
from .plugin_components import (
    AttrsConfigEditor,
    GroupBox,
    MiscErrorView,
    PidgetFactoryMapping,
    PidgetGroupFactoryMapping,
    SessionConfigEditor,
    SmartMetadataView,
    SmartPerfCalcView,
    TwoSensorIdsEditor,
    pidgets,
)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import json
import traceback
import typing as t
from pathlib import Path

import typing_extensions as te

from PySide6.QtCore import QSize, QTimer
from PySide6.QtWidgets import QFileDialog, QHBoxLayout, QPushButton, QWidget

from acconeer.exptool.app.new._exceptions import HandledException
from acconeer.exptool.app.new.ui.icons import CHECKMARK, FOLDER_OPEN, SAVE
from acconeer.exptool.app.new.ui.misc import ExceptionWidget

from .data_editor import DataEditor
from .save_dialog import (
    PresentationType,
    PresenterFunc,
    SaveDialogWithPreview,
    set_config_presenter,
)


@te.runtime_checkable
class JsonPresentable(te.Protocol):
    def to_json(self) -> str:
        ...

    @classmethod
    def from_json(self, json: str) -> JsonPresentable:
        ...


def _none_coalescing_chain(*funcs: PresenterFunc) -> PresenterFunc:
    """
    Composes a sequence of PresenterFuncs into a chain.

    The chain will call the functions in order,
    terminating at (and returning the result of)
    the first function returning a non-None value.
    """

    def _chain(instance: t.Any, t: PresentationType) -> t.Optional[str]:
        for f in funcs:
            presentation = f(instance, t)
            if presentation is not None:
                return presentation
        return None

    return _chain


def _json_presentation(instance: t.Any, t: PresentationType) -> t.Optional[str]:
    if t is PresentationType.JSON and isinstance(instance, JsonPresentable):
        return json.dumps(
            json.loads(instance.to_json()),
            indent=2,
        )

    return None


class JsonSaveLoadButtons(QWidget):
    _ICON_SIZE = QSize(20, 20)
    _BUTTON_SIZE = QSize(35, 25)
    _ICON_TRANSIENT_CHECKMARK_DURATION_MS = 2000
    _FILE_DIALOG_OPTIONS = dict(
        filter="JSON (*.json)",
        options=QFileDialog.Option.DontUseNativeDialog,
    )

    def __init__(
        self,
        getter: t.Callable[[], JsonPresentable],
        setter: t.Callable[[JsonPresentable], t.Any],
        encoder: t.Callable[[JsonPresentable], str],
        decoder: t.Callable[[str], JsonPresentable],
        extra_presenter: PresenterFunc,
        parent: t.Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)

        self._getter = getter
        self._setter = setter
        self._encoder = encoder
        self._decoder = decoder
        self._extra_presenter = extra_presenter

        self._save = QPushButton(SAVE(), "")
        self._save.setFixedSize(self._BUTTON_SIZE)
        self._save.setIconSize(self._ICON_SIZE)
        self._save.setToolTip("Save the current config to file")
        self._save.clicked.connect(
            self._except_errors(
                self._save_to_selected_file,
                header="Could not save config to file.",
            )
        )

        self._load = QPushButton(FOLDER_OPEN(), "")
        self._load.setFixedSize(self._BUTTON_SIZE)
        self._load.setIconSize(self._ICON_SIZE)
        self._load.setToolTip("Load a config from file")
        self._load.clicked.connect(
            self._except_errors(
                self._load_and_set_selected_file,
                header="Could not load config from file.",
            )
        )

        self.setLayout(QHBoxLayout())
        self.layout().setContentsMargins(5, 5, 5, 8)
        self.layout().addWidget(self._save)
        self.layout().addWidget(self._load)

    @classmethod
    def from_editor_and_config_type(
        cls,
        editor: DataEditor[JsonPresentable],
        config_type: t.Type[JsonPresentable],
        extra_presenter: PresenterFunc = lambda i, t: None,
    ) -> JsonSaveLoadButtons:
        """
        Creates a JsonSaveLoadButtons instance with a built-in JSON preview

        :param editor: The editor to bind the save/load buttons to
        :param config_type: The type (class) of the config
        :param extra_presenter:
            A PresenterFunc hooks into preview presentation creation.
            This function should always return None if its arguments aren't handled.
        """
        buttons = cls(
            editor.get_data,
            editor.sig_update.emit,
            config_type.to_json,
            config_type.from_json,
            extra_presenter,
        )
        editor.sig_update.connect(lambda model: buttons._save.setEnabled(model is not None))
        return buttons

    def _save_to_selected_file(self) -> None:
        model = self._getter()
        filename = SaveDialogWithPreview.get_save_file_name(
            caption="Save to file",
            model=model,
            presenter=_none_coalescing_chain(
                _json_presentation,
                set_config_presenter,
                self._extra_presenter,
            ),
            **self._FILE_DIALOG_OPTIONS,
        )

        if filename:
            Path(filename).with_suffix(".json").write_text(self._encoder(model))
            self._show_transient_checkmark(self._save)

    def _load_and_set_selected_file(self) -> None:
        filename, _ = QFileDialog.getOpenFileName(
            caption="Load from file", **self._FILE_DIALOG_OPTIONS
        )

        if filename:
            self._setter(self._decoder(Path(filename).read_text()))
            self._show_transient_checkmark(self._load)

    @classmethod
    def _show_transient_checkmark(cls, button: QPushButton) -> None:
        original_icon = button.icon()
        button.setIcon(CHECKMARK())
        QTimer.singleShot(
            cls._ICON_TRANSIENT_CHECKMARK_DURATION_MS, lambda: button.setIcon(original_icon)
        )

    def _except_errors(self, f: t.Callable[..., t.Any], header: str) -> t.Callable[..., t.Any]:
        def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            try:
                return f(*args, **kwargs)
            except Exception as e:
                ExceptionWidget(
                    parent=self,
                    exc=HandledException(
                        header + "\n" + f"The following exception was raised: {e!r}\n"
                    ),
                    traceback_str=traceback.format_exc(),
                ).exec()

        return wrapper



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import enum
from typing import List, Optional

import attrs
import numpy as np
import numpy.typing as npt
from scipy.signal import butter

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import (
    ENVELOPE_FWHM_M,
    AlgoParamEnum,
    AlgoProcessorConfigBase,
    ProcessorBase,
    exponential_smoothing_coefficient,
)
from acconeer.exptool.a121.algo.presence import Processor as PresenceProcessor
from acconeer.exptool.a121.algo.presence import ProcessorConfig as PresenceProcessorConfig
from acconeer.exptool.a121.algo.presence import ProcessorResult as PresenceProcessorResult


class AppState(AlgoParamEnum):
    """Breathing app state."""

    INIT_STATE = enum.auto()

    NO_PRESENCE_DETECTED = enum.auto()
    """No presence detected."""

    INTRA_PRESENCE_DETECTED = enum.auto()
    """Intra presence detected."""

    DETERMINE_DISTANCE_ESTIMATE = enum.auto()
    """Determining distance using presence."""

    ESTIMATE_BREATHING_RATE = enum.auto()
    """Estimating breathing rate."""


@attrs.mutable(kw_only=True)
class BreathingProcessorConfig(AlgoProcessorConfigBase):
    lowest_breathing_rate: float = attrs.field(default=6.0)
    """Lowest anticipated breathing rate (breaths per minute)."""

    highest_breathing_rate: float = attrs.field(default=60.0)
    """Highest anticipated breathing rate (breaths per minute)."""

    time_series_length_s: float = attrs.field(default=20.0)
    """Time series length (s)."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> List[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if config.sensor_config.frame_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "frame_rate",
                    "Must be set",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class BreathingProcessorExtraResult:
    psd: npt.NDArray[np.float_]
    frequencies: npt.NDArray[np.float_]
    breathing_motion: npt.NDArray[np.float_]
    time_vector: npt.NDArray[np.float_]
    breathing_rate_history: npt.NDArray[np.float_]
    all_breathing_rate_history: npt.NDArray[np.float_]


@attrs.mutable(kw_only=True)
class BreathingProcessorResult:
    breathing_rate: Optional[float] = attrs.field(default=None)
    """Estimated breathing rate. Breaths per minute."""

    extra_result: BreathingProcessorExtraResult
    """Extra result, only used for visualization."""


class BreathingProcessor(ProcessorBase[BreathingProcessorConfig, BreathingProcessorResult]):
    """Breathing rate processor."""

    SECONDS_IN_MINUTE: float = 60.0
    HISTORY_S = SECONDS_IN_MINUTE * 2.0

    # Type declarations
    start_point: int
    end_point: int
    sparse_iq_buffer: npt.NDArray[np.complex_]
    filt_sparse_iq_buffer: npt.NDArray[np.complex_]
    angle_buffer: npt.NDArray[np.complex_]
    filt_angle_buffer: npt.NDArray[np.complex_]
    breathing_motion_buffer: npt.NDArray[np.float_]
    breathing_rate_history: npt.NDArray[np.float_]
    all_breathing_rate_history: npt.NDArray[np.float_]

    start_time: float
    init_counter: int
    point_counter: int
    prev_angle: Optional[float]
    lp_filt_ampl: Optional[float]
    angle_unwrapped: npt.NDArray[np.float_]

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        processor_config: BreathingProcessorConfig,
    ):
        assert sensor_config.frame_rate is not None

        lowest_breathing_rate_hz = processor_config.lowest_breathing_rate / self.SECONDS_IN_MINUTE
        highest_breathing_rate_hz = (
            processor_config.highest_breathing_rate / self.SECONDS_IN_MINUTE
        )
        self.frame_rate = sensor_config.frame_rate
        self.time_series_length_s = processor_config.time_series_length_s
        self.time_series_length = int(self.time_series_length_s * self.frame_rate)
        self.analysis_overlap = int(self.time_series_length / 2)
        self.num_points = sensor_config.num_points

        # Filter coefficients.
        self.b_static, self.a_static = butter(
            N=2, Wn=lowest_breathing_rate_hz, btype="lowpass", fs=self.frame_rate
        )
        self.b_angle, self.a_angle = butter(
            N=2,
            Wn=[lowest_breathing_rate_hz, highest_breathing_rate_hz],
            btype="bandpass",
            fs=self.frame_rate,
        )
        self.sf = exponential_smoothing_coefficient(self.frame_rate, self.time_series_length_s)

        # PSD frequency vector.
        self.frequencies = np.fft.rfftfreq(self.time_series_length, 1 / self.frame_rate)
        self.time_vector = np.linspace(-self.HISTORY_S, 0, int(self.frame_rate * self.HISTORY_S))

        self.reinitialize_processor(0, self.num_points)

    def process(self, result: a121.Result) -> BreathingProcessorResult:
        frame = result.frame[:, self.start_point : self.end_point]
        mean_sweep = frame.mean(axis=0)

        # Estimate static component
        self.sparse_iq_buffer = np.roll(self.sparse_iq_buffer, shift=1, axis=0)
        self.sparse_iq_buffer[0] = mean_sweep

        filt_sparse_iq = -np.sum(
            self.a_static[1:][:, np.newaxis] * self.filt_sparse_iq_buffer, axis=0
        ) + np.sum(self.b_static[:, np.newaxis] * self.sparse_iq_buffer, axis=0)

        self.filt_sparse_iq_buffer = np.roll(self.filt_sparse_iq_buffer, shift=1, axis=0)
        self.filt_sparse_iq_buffer[0] = filt_sparse_iq

        # Remove static components by subtracting the estimated mean.
        zm_sweep = mean_sweep - filt_sparse_iq
        angle = np.angle(zm_sweep)

        if self.prev_angle is None or self.lp_filt_ampl is None:
            self.prev_angle = angle
            self.lp_filt_ampl = np.abs(zm_sweep)

        # Filter the amplitude(used when weighting psds from the measured distances).
        assert self.lp_filt_ampl is not None
        self.lp_filt_ampl = self.sf * self.lp_filt_ampl + (1 - self.sf) * np.abs(zm_sweep)

        # Unwrap the angles.
        angle_diff = angle - self.prev_angle
        angle_diff[np.pi < angle_diff] -= 2 * np.pi
        angle_diff[angle_diff < -np.pi] += 2 * np.pi
        self.angle_unwrapped = self.angle_unwrapped + angle_diff
        self.angle_buffer = np.roll(self.angle_buffer, shift=1, axis=0)
        self.angle_buffer[0] = self.angle_unwrapped
        self.prev_angle = angle

        # Bandpass filter angles.
        filt_angle = -np.sum(
            self.a_angle[1:][:, np.newaxis] * self.filt_angle_buffer, axis=0
        ) + np.sum(self.b_angle[:, np.newaxis] * self.angle_buffer, axis=0)
        self.filt_angle_buffer = np.roll(self.filt_angle_buffer, shift=1, axis=0)
        self.filt_angle_buffer[0] = filt_angle

        # Add filtered angle to breathing motion fifo buffer.
        self.breathing_motion_buffer = np.roll(self.breathing_motion_buffer, shift=-1, axis=0)
        self.breathing_motion_buffer[-1] = filt_angle

        # Calculate psd of signal.
        windowed_breathing_motion_buffer = (
            self.breathing_motion_buffer * np.hamming(self.time_series_length)[:, np.newaxis]
        )
        psd = np.fft.rfft(windowed_breathing_motion_buffer, axis=0, n=self.time_series_length)
        # Omit **2 to reduce processing as it does not alter the result.
        psd = np.abs(psd)
        assert self.lp_filt_ampl is not None
        psd_weighted = np.sum(psd * self.lp_filt_ampl, axis=1) / np.sum(self.lp_filt_ampl)

        # Interpolate around peak to gain better resolution.
        # Wait until data of a full time series is available.
        peak_loc = np.argmax(psd_weighted)
        if peak_loc != 0 and self.time_series_length < self.init_counter:
            estimated_frequency = self._peak_interpolation(
                psd_weighted[peak_loc - 1 : peak_loc + 2],
                self.frequencies[peak_loc - 1 : peak_loc + 2],
            )
            estimated_breathing_rate = estimated_frequency * self.SECONDS_IN_MINUTE
        else:
            self.init_counter += 1
            estimated_breathing_rate = None

        # Shift breathing rate history and add latest estimate.
        self.all_breathing_rate_history = np.roll(self.all_breathing_rate_history, shift=-1)
        self.all_breathing_rate_history[-1] = estimated_breathing_rate
        self.breathing_rate_history = np.roll(self.breathing_rate_history, shift=-1)

        # Report breathing rate if enough time has elapsed since last estimate.
        if self.time_series_length - self.analysis_overlap <= self.point_counter:
            self.breathing_rate_history[-1] = estimated_breathing_rate
            self.point_counter = 0
        else:
            self.breathing_rate_history[-1] = np.nan
            self.point_counter += 1

        # Prepare extra result, used for plotting.
        extra_result = BreathingProcessorExtraResult(
            psd=psd_weighted,
            frequencies=self.frequencies,
            breathing_motion=self.breathing_motion_buffer[:, self.center_distance_idx],
            time_vector=self.time_vector,
            all_breathing_rate_history=self.all_breathing_rate_history,
            breathing_rate_history=self.breathing_rate_history,
        )

        return BreathingProcessorResult(
            breathing_rate=estimated_breathing_rate, extra_result=extra_result
        )

    def reinitialize_processor(self, start_point: int, end_point: int) -> None:
        self.start_point = start_point
        self.end_point = end_point

        num_points_to_analyze = self.end_point - self.start_point
        self.center_distance_idx = int(num_points_to_analyze / 2)

        # Memory of iir filters.
        self.sparse_iq_buffer = np.zeros(
            shape=(self.b_static.size, num_points_to_analyze), dtype="complex128"
        )
        self.filt_sparse_iq_buffer = np.zeros(
            shape=(self.a_static.size - 1, num_points_to_analyze), dtype="complex128"
        )

        self.angle_buffer = np.zeros(shape=(self.b_angle.size, num_points_to_analyze))
        self.filt_angle_buffer = np.zeros(shape=(self.a_angle.size - 1, num_points_to_analyze))

        # Memory for breathing motion time series.
        self.breathing_motion_buffer = np.zeros(
            shape=(self.time_series_length, num_points_to_analyze)
        )

        # State variables.
        self.init_counter = 0
        self.prev_angle = None
        self.lp_filt_ampl = None
        self.point_counter = 0
        self.angle_unwrapped = np.zeros(shape=num_points_to_analyze)

        # Memory for breathing rate history.
        self.breathing_rate_history = np.full(
            shape=int(self.frame_rate * self.HISTORY_S), fill_value=np.nan
        )
        self.all_breathing_rate_history = np.full(
            shape=int(self.frame_rate * self.HISTORY_S), fill_value=np.nan
        )

    @staticmethod
    def _peak_interpolation(y: npt.NDArray[np.float_], x: npt.NDArray[np.float_]) -> float:
        """Quadratic interpolation of three points.

        Derivation:
        https://math.stackexchange.com/questions/680646/get-polynomial-function-from-3-points
        """
        a = (x[0] * (y[2] - y[1]) + x[1] * (y[0] - y[2]) + x[2] * (y[1] - y[0])) / (
            (x[0] - x[1]) * (x[0] - x[2]) * (x[1] - x[2])
        )
        b = (y[1] - y[0]) / (x[1] - x[0]) - a * (x[0] + x[1])
        peak_loc = -b / (2 * a)
        return float(peak_loc)

    def update_config(self, config: BreathingProcessorConfig) -> None:
        raise NotImplementedError


def get_presence_config() -> PresenceProcessorConfig:
    presence_config = PresenceProcessorConfig()
    presence_config.intra_detection_threshold = 4.0
    return presence_config


class ProcessorConfig(AlgoProcessorConfigBase):
    num_distances_to_analyze: int = attrs.field(default=3)
    """Indicates the number of distance to analyzed, centered around the distance where presence
    is detected."""

    distance_determination_duration: float = attrs.field(default=5.0)
    """Time for the presence processor to determine distance to presence."""

    use_presence_processor: bool = attrs.field(default=True)
    """If True, use the presence processor to determine distance to subject."""

    breathing_config: BreathingProcessorConfig = attrs.field(factory=BreathingProcessorConfig)

    presence_config: PresenceProcessorConfig = attrs.field(factory=get_presence_config)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> List[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if config.sensor_config.frame_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "frame_rate",
                    "Must be set",
                )
            )

        return validation_results


@attrs.mutable(kw_only=True)
class ProcessorResult:
    app_state: AppState
    """Application state."""

    distances_being_analyzed: Optional[tuple[int, int]] = None
    """Range where breathing is being analyzed."""

    presence_result: PresenceProcessorResult
    """Presence processor result."""

    breathing_result: Optional[BreathingProcessorResult] = attrs.field(default=None)
    """Breathing processor result."""


class Processor(ProcessorBase[ProcessorConfig, ProcessorResult]):
    """Breathing rate super-processor.

    Handles execution of the presence processor and the breathing processor.
    """

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        processor_config: ProcessorConfig,
        metadata: a121.Metadata,
    ):
        assert sensor_config.frame_rate is not None

        self.sensor_config = sensor_config
        self.metadata = metadata

        self.presence_processor = PresenceProcessor(
            sensor_config=sensor_config,
            metadata=self.metadata,
            processor_config=processor_config.presence_config,
        )

        self.breathing_processor = BreathingProcessor(
            sensor_config=sensor_config,
            processor_config=processor_config.breathing_config,
        )

        self.breathing_processor.reinitialize_processor(0, sensor_config.num_points)

        # Parameters
        self.intra_detection_threshold = processor_config.presence_config.intra_detection_threshold
        self.use_presence_processor = processor_config.use_presence_processor
        self.presence_threshold = ENVELOPE_FWHM_M[sensor_config.profile] * 2.0
        self.num_distances_to_analyze_half_width = int(
            processor_config.num_distances_to_analyze / 2.0
        )
        self.distance_determination_duration = (
            processor_config.distance_determination_duration * sensor_config.frame_rate
        )
        self.sf = exponential_smoothing_coefficient(
            sensor_config.frame_rate, processor_config.distance_determination_duration / 4.0
        )
        self.num_points = sensor_config.num_points

        # State variables
        self.presence_distance_has_init = False
        self.distance_determination_counter = 0
        self.app_state: AppState = AppState.INIT_STATE
        self.prev_app_state = AppState.INIT_STATE
        self.base_presence_distance: Optional[float] = None

    def process(self, result: a121.Result) -> ProcessorResult:
        presence_result = self.presence_processor.process(result=result)

        self._determine_app_state(presence_result)

        self._update_presence_distance(presence_result.presence_distance)

        processor_result = ProcessorResult(
            app_state=self.app_state, presence_result=presence_result
        )

        self._perform_action_based_on_app_state(result, processor_result)

        self.prev_app_state = self.app_state

        return processor_result

    def _determine_app_state(self, presence_result: PresenceProcessorResult) -> None:
        if not presence_result.presence_detected:
            self.app_state = AppState.NO_PRESENCE_DETECTED

        elif self.intra_detection_threshold < presence_result.intra_presence_score:
            self.app_state = AppState.INTRA_PRESENCE_DETECTED

        elif self.base_presence_distance is None and self.use_presence_processor:
            self.app_state = AppState.DETERMINE_DISTANCE_ESTIMATE

        elif (
            not self.use_presence_processor
            or self.distance_determination_duration <= self.distance_determination_counter
        ):
            self.app_state = AppState.ESTIMATE_BREATHING_RATE

        else:
            # Do not change app state
            pass

    def _update_presence_distance(self, presence_distance: float) -> None:
        # Calculate and filter presence distance.
        if not self.presence_distance_has_init:
            self.presence_distance = presence_distance
            self.presence_distance_has_init = True

        assert self.presence_distance is not None
        self.presence_distance = self.presence_distance * self.sf + presence_distance * (
            1 - self.sf
        )

        # Determine if presence location has changed.
        if self.base_presence_distance is not None and self.presence_threshold < np.abs(
            self.base_presence_distance - self.presence_distance
        ):
            self.base_presence_distance = None

    def _perform_action_based_on_app_state(
        self, result: a121.Result, processor_result: ProcessorResult
    ) -> None:
        # Perform action based on app state
        if (
            self.app_state == AppState.INTRA_PRESENCE_DETECTED
            or self.app_state == AppState.NO_PRESENCE_DETECTED
        ):
            self.base_presence_distance = None

        elif self.app_state == AppState.DETERMINE_DISTANCE_ESTIMATE:

            if self.app_state != self.prev_app_state:
                self.distance_determination_counter = 0
            else:
                self.distance_determination_counter += 1
                self.base_presence_distance = self.presence_distance

        elif self.app_state == AppState.ESTIMATE_BREATHING_RATE:
            if self.app_state != self.prev_app_state:
                if self.use_presence_processor:
                    center_idx = self._base_presence_distance_to_point()
                    start_point = max(0, center_idx - self.num_distances_to_analyze_half_width)
                    end_point = min(
                        center_idx + self.num_distances_to_analyze_half_width + 1,
                        self.num_points,
                    )
                else:
                    start_point = 0
                    end_point = self.num_points

                self.breathing_processor.reinitialize_processor(
                    start_point=start_point,
                    end_point=end_point,
                )

            processor_result.distances_being_analyzed = (
                self.breathing_processor.start_point,
                self.breathing_processor.end_point,
            )
            processor_result.breathing_result = self.breathing_processor.process(result)

        else:
            raise NotImplementedError("Invalid app state")

    def _base_presence_distance_to_point(self) -> int:
        """Calculates the closest point of a distance in meters"""
        assert self.base_presence_distance is not None
        assert not isinstance(self.metadata, list)

        measured_points = (
            self.sensor_config.start_point
            + np.arange(self.sensor_config.num_points) * self.sensor_config.step_length
        )
        center_point = int(self.base_presence_distance / self.metadata.base_step_length_m)
        return int(np.argmin(np.abs(measured_points - center_point)))

    def update_config(self, config: ProcessorConfig) -> None:
        raise NotImplementedError



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import abc
from typing import Any, List, Optional, Type, Union

from acconeer.exptool.a121._core.entities import (
    ClientInfo,
    Metadata,
    Result,
    SensorCalibration,
    SensorConfig,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._core.mediators import Recorder

from .communication_protocol import CommunicationProtocol


class ClientError(Exception):
    pass


class ClientCreationError(Exception):
    pass


class ClientABCWithGoodError(abc.ABC):
    def __new__(cls, *args: Any, **kwargs: Any) -> ClientABCWithGoodError:
        try:
            return super().__new__(cls)  # , *args, **kwargs)
        except TypeError as te:  # te here is the "Can't instantiate ..."-error
            raise ClientCreationError("Client cannot be instantiated, use Client.open()") from te


class Client(ClientABCWithGoodError):

    __registry: List[Type[Client]] = []

    @classmethod
    def open(
        cls,
        ip_address: Optional[str] = None,
        tcp_port: Optional[int] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, bool]] = None,
        mock: Optional[bool] = None,
        override_baudrate: Optional[int] = None,
        _override_protocol: Optional[Type[CommunicationProtocol]] = None,
    ) -> Client:
        """
        Open a new client
        """
        if len([e for e in [ip_address, serial_port, usb_device, mock] if e is not None]) > 1:
            raise ValueError("Only one connection can be selected")

        client = None
        for subclass in cls.__registry:
            try:
                client = subclass.open(
                    ip_address,
                    tcp_port,
                    serial_port,
                    usb_device,
                    mock,
                    override_baudrate,
                    _override_protocol,
                )
            except ClientCreationError:
                pass

        if client is not None:
            client._open()
            return client

        # This should not happen since the current implementation
        # only has two clients which are mutual exclusive.
        # * The mock client (mock is not None)
        # * The exploration client (mock is None)
        raise ClientCreationError("No client could be created")

    @classmethod
    def _register(cls, subclass: Type[Client]) -> Type[Client]:
        """Registers a subclass"""
        if not issubclass(subclass, cls):
            raise TypeError(f"{subclass.__name__!r} needs to be a subclass of {cls.__name__}.")
        cls.__registry.append(subclass)
        return subclass

    @abc.abstractmethod
    def _open(self) -> None:
        """Connects to client, called from open"""
        ...

    @abc.abstractmethod
    def setup_session(
        self,
        config: Union[SensorConfig, SessionConfig],
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> Union[Metadata, list[dict[int, Metadata]]]:
        """Sets up the session specified by ``config``.

        :param config: The session to set up.
        :param calibrations: An optional dict with :class:`SensorCalibration` for the session.
        :raises:
            ``ValueError`` if the config is invalid.

        :returns:
            ``Metadata`` if ``config.extended is False``,
            ``list[dict[int, Metadata]]`` otherwise.
        """
        ...

    @abc.abstractmethod
    def start_session(self) -> None:
        """Starts the already set up session.

        After this call, the server starts streaming data to the client.

        :param recorder:
            An optional ``Recorder``, which samples every ``get_next()``
        :raises: ``ClientError`` if ``Client``'s  session is not set up.
        """
        ...

    @abc.abstractmethod
    def get_next(self) -> Union[Result, list[dict[int, Result]]]:
        """Gets results from the server.

        :returns:
            A ``Result`` if the setup ``SessionConfig.extended is False``,
            ``list[dict[int, Result]]`` otherwise.
        :raises:
            ``ClientError`` if ``Client``'s session is not started.
        """
        ...

    @abc.abstractmethod
    def stop_session(self) -> None:
        """Stops an on-going session

        :raises:
            ``ClientError`` if ``Client``'s session is not started.
        """
        ...

    @abc.abstractmethod
    def attach_recorder(self, recorder: Recorder) -> None:
        """Attaches a recorder to this client."""
        ...

    @abc.abstractmethod
    def detach_recorder(self) -> Optional[Recorder]:
        """Detaches and returns the attached recorder (if present)."""
        ...

    @abc.abstractmethod
    def close(self) -> None:
        """Closes the connection to the host"""
        ...

    @property
    @abc.abstractmethod
    def connected(self) -> bool:
        """Whether this Client is connected."""
        ...

    @property
    @abc.abstractmethod
    def session_is_setup(self) -> bool:
        """Whether this Client has a session set up."""
        ...

    @property
    @abc.abstractmethod
    def session_is_started(self) -> bool:
        """Whether this Client's session is started."""
        ...

    @property
    @abc.abstractmethod
    def server_info(self) -> ServerInfo:
        """The ``ServerInfo``."""
        ...

    @property
    @abc.abstractmethod
    def client_info(self) -> ClientInfo:
        """The ``ClientInfo``."""
        ...

    @property
    @abc.abstractmethod
    def session_config(self) -> SessionConfig:
        """The :class:`SessionConfig` for the current session"""
        ...

    @property
    @abc.abstractmethod
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        """The extended :class:`Metadata` for the current session"""
        ...

    @property
    @abc.abstractmethod
    def calibrations(self) -> dict[int, SensorCalibration]:
        """
        Returns a dict with a :class:`SensorCalibration` per used
        sensor for the current session:

        For example, if session_setup was called with

        .. code-block:: python

            client.setup_session(
                SessionConfig({1: SensorConfig(), 3: SensorConfig()}),
            )

        this attribute will return {1: SensorCalibration(...), 3: SensorCalibration(...)}
        """
        ...

    @property
    @abc.abstractmethod
    def calibrations_provided(self) -> dict[int, bool]:
        """
        Returns whether a calibration was provided for each sensor in
        setup_session. For example, if setup_session was called with

        .. code-block:: python

            client.setup_session(
                SessionConfig({1: SensorConfig(), 2: SensorConfig()}),
                calibrations={2: SensorCalibration(...)},
            )

        this attribute will return ``{1: False, 2: True}``
        """
        ...

    def __enter__(self) -> Client:
        return self

    def __exit__(self, *_: Any) -> None:
        self.close()



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import enum
import time
from typing import Dict, List, Optional

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import AlgoProcessorConfigBase
from acconeer.exptool.a121.algo.distance import DetectorResult


TIME_HISTORY_S = 30
UPDATES_PER_SECOND = 4


class ProcessorLevelStatus(enum.Enum):
    """Status from the processor"""

    IN_RANGE = enum.auto()
    NO_DETECTION = enum.auto()
    OVERFLOW = enum.auto()
    OUT_OF_RANGE = enum.auto()


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    level_and_time_for_plotting: dict[str, npt.NDArray[np.float_]]


@attrs.frozen(kw_only=True)
class ProcessorResult:
    """Processor results"""

    peak_detected: Optional[bool]
    peak_status: Optional[ProcessorLevelStatus]
    filtered_level: Optional[float]
    extra_result: ProcessorExtraResult


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    median_filter_length: int = attrs.field(default=5)
    """Length of the median filter used to stabilize the level result."""
    num_medians_to_average: int = attrs.field(default=5)
    """Number of median values averaged to obtain the level result."""
    tank_range_start_m: float = attrs.field(default=0.030)
    """Minimum detection distance."""
    tank_range_end_m: float = attrs.field(default=0.50)
    """Maximum detection distance."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        return []


class Processor:
    """Tank level indicator processor
    :param processor_config: Processor configuration.
    """

    def __init__(
        self,
        processor_config: ProcessorConfig,
    ):
        self.tank_range_end_m = processor_config.tank_range_end_m
        self.tank_range_start_m = processor_config.tank_range_start_m
        self.median_filter_length = processor_config.median_filter_length
        self.num_medians_to_average = processor_config.num_medians_to_average

        assert self.tank_range_start_m < self.tank_range_end_m

        self.median_counter = 0
        self.mean_counter = 0

        self.level_history = np.full(self.median_filter_length, np.nan)
        self.median_vector = np.full(self.num_medians_to_average, np.nan)

        self.near_edge_status_list: List[Optional[bool]] = []
        self.peak_status_list: List[bool] = []

        self.level_and_time_for_plotting = {
            "time": np.full((TIME_HISTORY_S * UPDATES_PER_SECOND,), np.nan),
            "level": np.full((TIME_HISTORY_S * UPDATES_PER_SECOND,), np.nan),
        }

        self.start_time = time.time()

    def _update_level_history(self, result: DetectorResult) -> None:
        self.median_counter += 1
        assert result.distances is not None

        if len(result.distances) != 0:
            self.level_history = np.append(
                self.level_history, (self.tank_range_end_m - result.distances[0])
            )
        else:
            self.level_history = np.append(self.level_history, np.nan)
        self.level_history = self.level_history[1:]

        self.near_edge_status_list.append(result.near_edge_status)

        if self.median_counter == self.median_filter_length:
            self.median_counter = 0
            # store level in a vector of medians
            self.median_vector = np.append(self.median_vector, np.median(self.level_history))[1:]
            # store peak status from near_edge_status_list to peak_status_list
            self.peak_status_list.append(
                self.near_edge_status_list.count(True) > len(self.near_edge_status_list) / 2
            )
            # clear the edge status list
            self.near_edge_status_list.clear()
            self.mean_counter += 1

    def _update_level_and_time_for_plotting(
        self, filtered_level: Optional[float], rel_time: float
    ) -> None:
        if filtered_level is not None:
            self.level_and_time_for_plotting["time"] = (
                np.append(self.level_and_time_for_plotting["time"], 0) - rel_time
            )[1:]
            self.level_and_time_for_plotting["level"] = (
                np.append(self.level_and_time_for_plotting["level"], filtered_level)
            )[1:]

    def process(self, detector_result: Dict[int, DetectorResult]) -> ProcessorResult:
        # Get the first detector result (single sensor operation).
        (result,) = list(detector_result.values())

        # Assign level status
        # peak_detected: True if a peak is detected by the distance detector in n consecutive
        # frames, where n=median_filter_length.
        # peak_status: Assigned based on the location of the peak in relation to the detection
        # range defined by tank_start and tank_end.
        self._update_level_history(result)

        peak_detected = None
        peak_status = None
        filtered_level = None

        if self.mean_counter == self.num_medians_to_average:
            self.mean_counter = 0
            # assign filtered_level every num_medians_to_average samples
            if np.any(~np.isnan(self.median_vector)):
                filtered_level = np.nanmean(self.median_vector)
            else:
                filtered_level = np.nan

            # assign the detection and peak status
            peak_detected = False
            peak_status = ProcessorLevelStatus.NO_DETECTION

            if ~np.isnan(filtered_level):
                peak_detected = True
                if filtered_level < 0:
                    peak_status = ProcessorLevelStatus.OUT_OF_RANGE
                elif filtered_level > self.tank_range_end_m - self.tank_range_start_m:
                    peak_status = ProcessorLevelStatus.OVERFLOW
                elif (
                    filtered_level <= self.tank_range_end_m - self.tank_range_start_m
                    and self.peak_status_list.count(True) > len(self.peak_status_list) / 2
                ):
                    peak_status = ProcessorLevelStatus.OVERFLOW
                else:
                    peak_status = ProcessorLevelStatus.IN_RANGE
            else:
                if self.peak_status_list.count(True) > len(self.peak_status_list) / 2:
                    peak_status = ProcessorLevelStatus.OVERFLOW

            if peak_status in (ProcessorLevelStatus.OVERFLOW, ProcessorLevelStatus.OUT_OF_RANGE):
                filtered_level = np.nan

            rel_time = time.time() - self.start_time
            if rel_time >= 1 / UPDATES_PER_SECOND:
                self._update_level_and_time_for_plotting(filtered_level, rel_time)
                self.start_time = time.time()

            # clear the peak status list
            self.peak_status_list.clear()

        # extra results for plotting
        extra_result = ProcessorExtraResult(
            level_and_time_for_plotting=self.level_and_time_for_plotting
        )

        return ProcessorResult(
            peak_detected=peak_detected,
            peak_status=peak_status,
            extra_result=extra_result,
            filtered_level=filtered_level,
        )



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import datetime
import os
from pathlib import Path
from time import time
from typing import Any, Optional, TypeVar
from uuid import uuid4

import h5py
import numpy as np

import acconeer.exptool
from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121._core.entities import (
    INT_16_COMPLEX,
    ClientInfo,
    Metadata,
    Result,
    SensorCalibration,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._core.mediators import Recorder, RecorderAttachable
from acconeer.exptool.utils import get_module_version  # type: ignore[import]

from .session_schema import SessionSchema
from .utils import PathOrH5File, h5_file_factory


T = TypeVar("T")


def get_h5py_str_dtype() -> Any:
    return h5py.special_dtype(vlen=str)


_H5PY_STR_DTYPE = get_h5py_str_dtype()


def get_timestamp() -> str:
    return datetime.datetime.now().isoformat(timespec="seconds")


def get_uuid() -> str:
    return str(uuid4())


class H5Recorder(Recorder):
    """Recorder writing directly to an HDF5 file

    :param path_or_file:
        If a path-like object, by default (depending on ``mode`` below) an HDF5 file is created at
        that path. When stopping, the file is closed.

        If an ``h5py.File``, that file is used as-is. In this case, the file is not closed when
        stopping.
    :param attachable:
        A ``Client`` that the recorder will be attached to. When a recorder is attached to a
        ``Client``, all metadata and data frames will be recorded.
    :param mode:
        The file mode to use if a path-like object was given for ``path_or_file``. Default value is
        'x', meaning that we open for exclusive creation, failing if the file already exists.
    :param _chunk_size:
        If given, data will be written to file every ``_chunk_size`` samples.

        If not given, data will be written at least every 512:th sample, or at least once per
        second, whichever comes first.

        Setting a small chunk size (e.g. 1) may degrade performance for high sample rates.

        Internal parameter, subject to change.
    """

    _AUTO_CHUNK_MAX_SIZE = 512
    _AUTO_CHUNK_MAX_TIME = 1.0
    _schema = SessionSchema

    path: Optional[os.PathLike[Any]]
    """The file path, if a path-like object was given for ``path_or_file``."""
    file: h5py.File
    """The ``h5py.File``."""
    owns_file: bool
    """Whether :class:`H5Recorder` opened and owns the file, i.e., if a path-like object was given
    for ``path_or_file``. If it does, the file is closed when stopping.
    """
    _num_frames_current_session: int
    _chunk_size: Optional[int]
    _chunk_buffer: list[list[dict[int, Result]]]
    _last_write_time: float

    def __init__(
        self,
        path_or_file: PathOrH5File,
        attachable: Optional[RecorderAttachable] = None,
        mode: str = "x",
        *,
        _chunk_size: Optional[int] = None,
        _lib_version: Optional[str] = None,
        _timestamp: Optional[str] = None,
        _uuid: Optional[str] = None,
    ) -> None:
        self.file, self.owns_file = h5_file_factory(path_or_file, h5_file_mode=mode)
        self.path = Path(self.file.filename) if self.owns_file else None
        self._chunk_size = _chunk_size
        self._chunk_buffer = []
        self._num_frames_current_session = 0
        self._last_write_time = 0.0
        self._current_session_group: Optional[h5py.Group] = None

        if attachable is None:
            self._attachable = None
        else:
            self._attachable = attachable
            attachable.attach_recorder(self)

        if _lib_version is None:
            _lib_version = get_module_version(acconeer.exptool)

        if _timestamp is None:
            _timestamp = get_timestamp()

        if _uuid is None:
            _uuid = get_uuid()

        self.file.create_dataset(
            "uuid",
            data=_uuid,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "timestamp",
            data=_timestamp,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "lib_version",
            data=_lib_version,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "generation",
            data="a121",
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

    def _start(
        self,
        *,
        client_info: ClientInfo,
        server_info: ServerInfo,
    ) -> None:
        if "client_info" in self.file or "server_info" in self.file:
            raise RuntimeError("It's not allowed to call '_start' once.")

        self.file.create_dataset(
            "client_info",
            data=client_info.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "server_info",
            data=server_info.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self._last_write_time = time()

    def _start_session(
        self,
        *,
        session_config: SessionConfig,
        extended_metadata: list[dict[int, Metadata]],
        calibrations: Optional[dict[int, SensorCalibration]],
        calibrations_provided: Optional[dict[int, bool]],
    ) -> None:
        self._current_session_group = self._schema.create_next_session_group(self.file)

        self._current_session_group.create_dataset(
            "session_config",
            data=session_config.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

        for i, metadata_group_dict in enumerate(extended_metadata):
            group_group = self._current_session_group.create_group(f"group_{i}")

            for entry_id, (sensor_id, metadata) in enumerate(metadata_group_dict.items()):
                entry_group = group_group.create_group(f"entry_{entry_id}")
                entry_group.create_dataset("sensor_id", data=sensor_id, track_times=False)

                entry_group.create_dataset(
                    "metadata",
                    data=metadata.to_json(),
                    dtype=_H5PY_STR_DTYPE,
                    track_times=False,
                )

                result_group = entry_group.create_group("result")
                self._create_result_datasets(result_group, metadata)

        if (calibrations is None) != (calibrations_provided is None):
            raise ValueError(
                "'calibrations_provided' must be provided if 'calibrations' is provided"
            )

        if calibrations is not None and calibrations_provided is not None:
            calibrations_group = self._current_session_group.create_group("calibrations")
            for sensor_id, calibration in calibrations.items():
                sensor_calibration_group = calibrations_group.create_group(f"sensor_{sensor_id}")

                calibration.to_h5(sensor_calibration_group)

                sensor_calibration_group.create_dataset(
                    "provided", data=calibrations_provided[sensor_id], track_times=False
                )

        self._last_write_time = time()

    def _write_chunk_buffer_to_file(self, start_idx: int) -> int:
        """Saves the contents of ``self.chunk_buffer`` to file.

        :returns: the number of extended results saved.
        """
        if len(self._chunk_buffer) == 0:
            return 0

        if self._current_session_group is None:
            raise RuntimeError("No session group selected yet. This should not happen.")

        for group_idx, entry_idx, results in utils.iterate_extended_structure_as_entry_list(
            utils.transpose_extended_structures(self._chunk_buffer)
        ):
            self._write_results(
                g=self._current_session_group[f"group_{group_idx}/entry_{entry_idx}/result"],
                start_index=start_idx,
                results=results,
            )

        return len(self._chunk_buffer)

    def _sample(self, extended_result: list[dict[int, Result]]) -> None:
        self._chunk_buffer.append(extended_result)

        if self._chunk_size is None:
            reached_size_limit = len(self._chunk_buffer) >= self._AUTO_CHUNK_MAX_SIZE
            reached_time_limit = (time() - self._last_write_time) >= self._AUTO_CHUNK_MAX_TIME
            write = reached_size_limit or reached_time_limit
        else:
            write = len(self._chunk_buffer) == self._chunk_size

        if write:
            self._num_frames_current_session += self._write_chunk_buffer_to_file(
                start_idx=self._num_frames_current_session
            )
            self._chunk_buffer = []
            self._last_write_time = time()

    def _stop_session(self) -> None:
        if len(self._chunk_buffer) > 0:
            _ = self._write_chunk_buffer_to_file(start_idx=self._num_frames_current_session)
            self._chunk_buffer = []
        self._num_frames_current_session = 0

    def close(self) -> Any:
        try:
            self._stop_session()

            # If the client has self as an attached recorder, make sure to detatch it.
            if self._attachable is not None:
                _ = self._attachable.detach_recorder()
        finally:
            if self.owns_file:
                self.file.close()

    @staticmethod
    def _create_result_datasets(g: h5py.Group, metadata: Metadata) -> None:
        g.create_dataset(
            "data_saturated",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
            compression="gzip",
        )
        g.create_dataset(
            "frame_delayed",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
            compression="gzip",
        )
        g.create_dataset(
            "calibration_needed",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
            compression="gzip",
        )
        g.create_dataset(
            "temperature",
            shape=(0,),
            maxshape=(None,),
            dtype=int,
            track_times=False,
            compression="gzip",
        )

        g.create_dataset(
            "tick",
            shape=(0,),
            maxshape=(None,),
            dtype=np.dtype("int64"),
            track_times=False,
            compression="gzip",
        )

        g.create_dataset(
            "frame",
            shape=(0, *metadata.frame_shape),
            maxshape=(None, *metadata.frame_shape),
            dtype=INT_16_COMPLEX,
            track_times=False,
            compression="gzip",
        )

    @staticmethod
    def _write_results(g: h5py.Group, start_index: int, results: list[Result]) -> None:
        """Extends the Dataset to the appropriate (new) size with .resize,
        and then copies the data over
        """
        datasets_to_extend = [
            "data_saturated",
            "frame_delayed",
            "calibration_needed",
            "temperature",
            "tick",
            "frame",
        ]
        for dataset_name in datasets_to_extend:
            g[dataset_name].resize(size=start_index + len(results), axis=0)

        dataset_slice = slice(start_index, start_index + len(results))

        g["data_saturated"][dataset_slice] = [result.data_saturated for result in results]
        g["frame_delayed"][dataset_slice] = [result.frame_delayed for result in results]
        g["calibration_needed"][dataset_slice] = [result.calibration_needed for result in results]
        g["temperature"][dataset_slice] = [result.temperature for result in results]
        g["tick"][dataset_slice] = [result.tick for result in results]
        g["frame"][dataset_slice] = [result._frame for result in results]

    def require_algo_group(self, key: str) -> h5py.Group:
        """Creates/gets the ``algo`` group with a given ``key``.

        :raises: Exception if the key does not match the file content
        """

        group = self.file.require_group("algo")

        if "key" in group:
            existing_key = bytes(group["key"][()]).decode()
            if existing_key != key:
                raise Exception(f"Algo group key mismatch: got '{key}' but had '{existing_key}'")
        else:
            group.create_dataset(
                "key",
                data=key,
                dtype=_H5PY_STR_DTYPE,
                track_times=False,
            )

        return group



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import typing as t

import h5py
import numpy as np

from acconeer.exptool.utils import PhonySeries  # type: ignore[import]

from ._processor import ProcessorLevelStatus
from ._ref_app import RefAppResult


_ALL_REF_APP_RESULT_FIELDS = (
    "peak_detected",
    "peak_status",
    "level",
    # "extra_result" ignored by default
)


class RefAppResultListH5Serializer:
    """
    Reads or writes a tank level RefAppResult from/to a given h5py.Group
    """

    def __init__(
        self,
        group: h5py.Group,
        fields: t.Sequence[str] = _ALL_REF_APP_RESULT_FIELDS,
        allow_missing_fields: bool = False,
    ) -> None:
        """
        :param group: The H5 group that will have datasets read/written to it
        :param fields: The fields to serialize. Has no effect on deserialization
        :param allow_missing_fields:
            Whether it's acceptable to break type safety during de-serialization
            by replacing missing fields by ``None``. Has no effect on serialization.
        """
        self.group = group
        self.fields = fields
        self.allow_missing_fields = allow_missing_fields

    def serialize(self, results: t.List[RefAppResult]) -> None:
        if "extra_result" in self.fields:
            raise NotImplementedError(
                "'extra_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        result_ready = np.array(
            [
                r.peak_detected is not None and r.peak_status is not None and r.level is not None
                for r in results
            ]
        )

        self.group.create_dataset("result_ready", dtype=bool, data=result_ready, track_times=False)

        if "peak_detected" in self.fields:
            data = np.array(
                [r.peak_detected if r.peak_detected is not None else False for r in results]
            )

            self.group.create_dataset("peak_detected", dtype=bool, data=data, track_times=False)

        if "peak_status" in self.fields:
            data = np.array(
                [
                    r.peak_status.name.encode("utf-8")
                    if r.peak_status is not None
                    else ProcessorLevelStatus.NO_DETECTION.name.encode("utf-8")
                    for r in results
                ]
            )

            self.group.create_dataset(
                "peak_status", dtype=h5py.special_dtype(vlen=bytes), data=data, track_times=False
            )

        if "level" in self.fields:
            data = np.array([r.level if r.level is not None else np.nan for r in results])

            self.group.create_dataset("level", dtype=float, data=data, track_times=False)

    def deserialize(self, _: None) -> t.List[RefAppResult]:
        if "extra_result" in self.fields:
            raise NotImplementedError(
                "'extra_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        groups = (
            self.group.get("peak_detected", PhonySeries(None)),
            self.group.get("peak_status", PhonySeries(None)),
            self.group.get("level", PhonySeries(None)),
            self.group.get("result_ready", PhonySeries(None)),
        )

        if any(isinstance(g, PhonySeries) for g in groups) and not self.allow_missing_fields:
            raise ValueError("Some fields are missing while 'allow_missing_fields' is False.")

        if all(g is None for g in groups):
            return []

        return [
            RefAppResult(
                peak_detected=peak_detected if result_ready else None,
                peak_status=ProcessorLevelStatus[peak_status.decode("utf-8")]
                if peak_status is not None and result_ready
                else None,
                level=level if result_ready else None,
                extra_result=None,  # type: ignore[arg-type]
            )
            for (
                peak_detected,
                peak_status,
                level,
                result_ready,
            ) in zip(*groups)
        ]



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import typing as t

import h5py
import numpy as np

from acconeer.exptool.a121.algo.presence import (
    ProcessorExtraResult as PresenceProcessorExtraResult,
)
from acconeer.exptool.a121.algo.presence import ProcessorResult as PresenceProcessorResult
from acconeer.exptool.utils import PhonySeries  # type: ignore[import]

from ._processor import AppState, BreathingProcessorExtraResult, BreathingProcessorResult
from ._ref_app import RefAppResult


_ALL_REF_APP_RESULT_FIELDS = (
    "app_state",
    "distances_being_analyzed",
    # "presence_result", ignored by default
    "breathing_result",
)


class RefAppResultListH5Serializer:
    """
    Reads or writes a tank level RefAppResult from/to a given h5py.Group
    """

    def __init__(
        self,
        group: h5py.Group,
        fields: t.Sequence[str] = _ALL_REF_APP_RESULT_FIELDS,
        allow_missing_fields: bool = False,
    ) -> None:
        """
        :param group: The H5 group that will have datasets read/written to it
        :param fields: The fields to serialize. Has no effect on deserialization
        :param allow_missing_fields:
            Whether it's acceptable to break type safety during deserialization
            by replacing missing fields by ``None``. Has no effect on serialization.
        """
        self.group = group
        self.fields = fields
        self.allow_missing_fields = allow_missing_fields

    def serialize(self, results: t.List[RefAppResult]) -> None:
        if "presence_result" in self.fields:
            raise NotImplementedError(
                "'presence_result' is not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        if "app_state" in self.fields:
            data = np.array([r.app_state.name.encode("utf-8") for r in results])

            self.group.create_dataset(
                "app_state", dtype=h5py.special_dtype(vlen=bytes), data=data, track_times=False
            )

        if "distances_being_analyzed" in self.fields:
            data = np.array(
                [
                    r.distances_being_analyzed
                    if r.distances_being_analyzed is not None
                    else [np.nan, np.nan]
                    for r in results
                ]
            )

            self.group.create_dataset(
                "distances_being_analyzed", dtype=float, data=data, track_times=False
            )

        if "breathing_result" in self.fields:
            data = np.array(
                [
                    r.breathing_result.breathing_rate
                    if r.breathing_result is not None
                    and r.breathing_result.breathing_rate is not None
                    else np.nan
                    for r in results
                ]
            )

            self.group.create_dataset(
                "breathing_result", dtype=float, data=data, track_times=False
            )

    def deserialize(self, _: None) -> t.List[RefAppResult]:
        if "presence_result" in self.fields:
            raise NotImplementedError(
                "'presence_result' is not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        groups = (
            self.group.get("app_state", PhonySeries(None)),
            self.group.get("distances_being_analyzed", PhonySeries(None)),
            self.group.get("breathing_result", PhonySeries(None)),
        )

        if any(isinstance(g, PhonySeries) for g in groups) and not self.allow_missing_fields:
            raise ValueError("Some fields are missing while 'allow_missing_fields' is False.")

        if all(g is None for g in groups):
            return []

        return [
            RefAppResult(
                app_state=AppState[app_state.decode("utf-8")],
                distances_being_analyzed=tuple(  # type: ignore
                    [int(d) for d in distances_being_analyzed]
                )
                if not np.isnan(distances_being_analyzed[0])
                else None,
                presence_result=PresenceProcessorResult(
                    intra_presence_score=0.0,
                    intra=np.array([], dtype=float),
                    inter_presence_score=0.0,
                    inter=np.array([], dtype=float),
                    presence_distance=0.0,
                    presence_detected=False,
                    extra_result=PresenceProcessorExtraResult(
                        frame=np.array([], dtype=complex),
                        abs_mean_sweep=np.array([], dtype=float),
                        fast_lp_mean_sweep=np.array([], dtype=float),
                        slow_lp_mean_sweep=np.array([], dtype=float),
                        lp_noise=np.array([], dtype=float),
                        presence_distance_index=0,
                    ),
                ),
                breathing_result=None
                if AppState[app_state.decode("utf-8")] != AppState.ESTIMATE_BREATHING_RATE
                else BreathingProcessorResult(
                    breathing_rate=None,
                    extra_result=BreathingProcessorExtraResult(
                        psd=np.array([], dtype=float),
                        frequencies=np.array([], dtype=float),
                        breathing_motion=np.array([], dtype=float),
                        time_vector=np.array([], dtype=float),
                        breathing_rate_history=np.array([], dtype=float),
                        all_breathing_rate_history=np.array([], dtype=float),
                    ),
                )
                if np.isnan(breathing_result)
                else BreathingProcessorResult(
                    breathing_rate=breathing_result,
                    extra_result=BreathingProcessorExtraResult(
                        psd=np.array([], dtype=float),
                        frequencies=np.array([], dtype=float),
                        breathing_motion=np.array([], dtype=float),
                        time_vector=np.array([], dtype=float),
                        breathing_rate_history=np.array([], dtype=float),
                        all_breathing_rate_history=np.array([], dtype=float),
                    ),
                ),
            )
            for (
                app_state,
                distances_being_analyzed,
                breathing_result,
            ) in zip(*groups)
        ]



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Tuple, Type

import numpy as np
import numpy.typing as npt

from PySide6.QtGui import QTransform
from PySide6.QtWidgets import QWidget

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121 import algo
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorPlotPluginBase,
    ProcessorPluginPreset,
    ProcessorPluginSpec,
    ProcessorViewPluginBase,
)
from acconeer.exptool.app.new import (
    AppModel,
    Message,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    pidgets,
)

from ._processor import (
    AmplitudeMethod,
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_sensor_config,
)


log = logging.getLogger(__name__)


class PluginPresetId(Enum):
    DEFAULT = auto()


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):

    PLUGIN_PRESETS = {
        PluginPresetId.DEFAULT.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_sensor_config()),
            processor_config=BackendPlugin.get_processor_config_cls()(),
        ),
    }

    @classmethod
    def get_processor_cls(cls) -> Type[Processor]:
        return Processor

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "amplitude_method": pidgets.EnumPidgetFactory(
                enum_type=AmplitudeMethod,
                name_label_text="Amplitude method:",
                name_label_tooltip=(
                    "The method used to calculate the amplitude from the complex Sparse IQ data"
                ),
                label_mapping={
                    AmplitudeMethod.COHERENT: "Coherent",
                    AmplitudeMethod.NONCOHERENT: "Non-coherent",
                    AmplitudeMethod.FFT_MAX: "FFT Max",
                },
            )
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def supports_multiple_subsweeps(self) -> bool:
        return True


class PlotPlugin(ProcessorPlotPluginBase[ProcessorResult]):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)
        self.smooth_max = et.utils.SmoothMax()

    def setup(self, metadata: a121.Metadata, sensor_config: a121.SensorConfig) -> None:
        self.ampl_plot = self._create_amplitude_plot(self.plot_layout, sensor_config.num_subsweeps)
        self.plot_layout.nextRow()
        self.phase_plot = self._create_phase_plot(self.plot_layout, sensor_config.num_subsweeps)
        self.plot_layout.nextRow()

        self.ampl_curves = []
        self.phase_curves = []
        self.ft_plots = []
        self.ft_im_list = []
        self.distances_m_list = []

        vels, vel_res = algo.get_approx_fft_vels(metadata, sensor_config)

        for i, subsweep in enumerate(sensor_config.subsweeps):
            distances_m, step_length_m = algo.get_distances_m(subsweep, metadata)
            self.distances_m_list.append(distances_m)

            ampl_curve = self._create_amplitude_curve(i, distances_m)
            self.ampl_plot.addItem(ampl_curve)
            self.ampl_curves.append(ampl_curve)

            phase_curve = self._create_phase_curve(i)
            self.phase_plot.addItem(phase_curve)
            self.phase_curves.append(phase_curve)

            ft_plot, ft_im = self._create_fft_plot(
                self.plot_layout,
                distances_m=distances_m,
                step_length_m=step_length_m,
                vels=vels,
                vel_res=vel_res,
            )
            self.ft_plots.append(ft_plot)
            self.ft_im_list.append(ft_im)

    def update(self, processor_result: ProcessorResult) -> None:
        max_ = 0.0

        for i, result in enumerate(processor_result):
            ampls = result.amplitudes
            self.ampl_curves[i].setData(self.distances_m_list[i], ampls)
            self.phase_curves[i].setData(self.distances_m_list[i], result.phases)
            dvm = result.distance_velocity_map
            self.ft_im_list[i].updateImage(
                dvm.T,
                levels=(0, 1.05 * np.max(dvm)),
            )

            max_ = max(max_, np.max(ampls).item())

        self.ampl_plot.setYRange(0, self.smooth_max.update(max_))

    @staticmethod
    def _create_amplitude_curve(
        cycle_num: int, depths_m: npt.NDArray[np.float_]
    ) -> pg.PlotDataItem:
        pen = et.utils.pg_pen_cycler(cycle_num)

        if len(depths_m) > 32:
            return pg.PlotDataItem(pen=pen)
        else:
            brush = et.utils.pg_brush_cycler(cycle_num)
            return pg.PlotDataItem(
                pen=pen, symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k"
            )

    @staticmethod
    def _create_phase_curve(cycle_num: int) -> pg.PlotDataItem:
        brush = et.utils.pg_brush_cycler(cycle_num)
        return pg.PlotDataItem(
            pen=None, symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k"
        )

    @staticmethod
    def _create_amplitude_plot(parent: pg.GraphicsLayout, colspan: int) -> pg.PlotItem:
        ampl_plot = parent.addPlot(colspan=colspan)
        ampl_plot.setMenuEnabled(False)
        ampl_plot.showGrid(x=True, y=True)
        ampl_plot.setLabel("left", "Amplitude")
        return ampl_plot

    @staticmethod
    def _create_phase_plot(parent: pg.GraphicsLayout, colspan: int) -> pg.PlotItem:
        phase_plot = parent.addPlot(colspan=colspan)
        phase_plot.setMenuEnabled(False)
        phase_plot.showGrid(x=True, y=True)
        phase_plot.setLabel("left", "Phase")
        phase_plot.setYRange(-np.pi, np.pi)
        phase_plot.getAxis("left").setTicks(et.utils.pg_phase_ticks)
        return phase_plot

    @staticmethod
    def _create_fft_plot(
        parent: pg.GraphicsLayout,
        *,
        distances_m: npt.NDArray[np.float_],
        step_length_m: float,
        vels: npt.NDArray[np.float_],
        vel_res: float,
    ) -> Tuple[pg.PlotItem, pg.ImageItem]:
        transform = QTransform()
        transform.translate(distances_m[0], vels[0] - 0.5 * vel_res)
        transform.scale(step_length_m, vel_res)

        im = pg.ImageItem(autoDownsample=True)
        im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
        im.setTransform(transform)

        plot = parent.addPlot()
        plot.setMenuEnabled(False)
        plot.setLabel("bottom", "Distance (m)")
        plot.setLabel("left", "Velocity (m/s)")
        plot.addItem(im)

        return plot, im


class PluginSpec(
    ProcessorPluginSpec[a121.Result, ProcessorConfig, ProcessorResult, a121.Metadata]
):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel, view_widget: QWidget) -> ViewPlugin:
        return ViewPlugin(app_model=app_model, view_widget=view_widget)

    def create_plot_plugin(
        self, app_model: AppModel, plot_layout: pg.GraphicsLayout
    ) -> PlotPlugin:
        return PlotPlugin(app_model=app_model, plot_layout=plot_layout)


SPARSE_IQ_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="sparse_iq",
    title="Sparse IQ",
    description="Basic usage of the sparse IQ service.",
    family=PluginFamily.SERVICE,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.smart_presence import RefApp, RefAppConfig


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client.open(**a121.get_client_args(args))

    ref_app_config = RefAppConfig(
        start_m=1.0,
        end_m=3.0,
        num_zones=3,
    )

    ref_app = RefApp(client=client, sensor_id=1, ref_app_config=ref_app_config)
    ref_app.start()

    pg_updater = PGUpdater(
        ref_app_config,
        ref_app.detector._get_sensor_config(ref_app_config),
        ref_app.detector.estimated_frame_rate,
        ref_app.ref_app_processor.zone_limits,
    )
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        ref_app_result = ref_app.get_next()
        if ref_app_result.presence_detected:
            print(f"Presence in zone {ref_app_result.max_presence_zone}")
        else:
            print("No presence")
        try:
            pg_process.put_data(ref_app_result)
        except et.PGProccessDiedException:
            break

    ref_app.stop()

    print("Disconnecting...")
    client.close()


class PGUpdater:
    def __init__(
        self,
        ref_app_config: RefAppConfig,
        sensor_config: a121.SensorConfig,
        estimated_frame_rate: float,
        zone_limits: npt.NDArray[np.float_],
    ):
        self.ref_app_config = ref_app_config

        self.history_length_s = 5
        self.estimated_frame_rate = estimated_frame_rate
        self.history_length_n = int(round(self.history_length_s * estimated_frame_rate))
        self.intra_history = np.zeros(self.history_length_n)
        self.inter_history = np.zeros(self.history_length_n)

        self.num_sectors = min(ref_app_config.num_zones, sensor_config.num_points)
        self.sector_size = max(1, -(-sensor_config.num_points // self.num_sectors))

        self.sector_offset = (self.num_sectors * self.sector_size - sensor_config.num_points) // 2
        self.zone_limits = zone_limits

        self.setup_is_done = False

    def setup(self, win):
        win.setWindowTitle("Acconeer smart presence example")

        self.intra_limit_lines = []
        self.inter_limit_lines = []

        # Intra presence history plot

        self.intra_hist_plot = win.addPlot(
            row=0,
            col=0,
            title="Intra presence history (fast motions)",
        )
        self.intra_hist_plot.setMenuEnabled(False)
        self.intra_hist_plot.setMouseEnabled(x=False, y=False)
        self.intra_hist_plot.hideButtons()
        self.intra_hist_plot.showGrid(x=True, y=True)
        self.intra_hist_plot.setLabel("bottom", "Time (s)")
        self.intra_hist_plot.setLabel("left", "Score")
        self.intra_hist_plot.setXRange(-self.history_length_s, 0)
        self.intra_history_smooth_max = et.utils.SmoothMax(self.estimated_frame_rate)
        self.intra_hist_plot.setYRange(0, 10)
        if not self.ref_app_config.intra_enable:
            intra_color = et.utils.color_cycler(1)
            intra_color = f"{intra_color}50"
            intra_dashed_pen = pg.mkPen(intra_color, width=2.5, style=QtCore.Qt.DashLine)
            intra_pen = pg.mkPen(intra_color, width=2)
        else:
            intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
            intra_pen = et.utils.pg_pen_cycler(1)

        self.intra_hist_curve = self.intra_hist_plot.plot(pen=intra_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=intra_dashed_pen)
        self.intra_hist_plot.addItem(limit_line)
        self.intra_limit_lines.append(limit_line)

        for line in self.intra_limit_lines:
            line.setPos(self.ref_app_config.intra_detection_threshold)

        # Inter presence history plot

        self.inter_hist_plot = win.addPlot(
            row=0,
            col=1,
            title="Inter presence history (slow motions)",
        )
        self.inter_hist_plot.setMenuEnabled(False)
        self.inter_hist_plot.setMouseEnabled(x=False, y=False)
        self.inter_hist_plot.hideButtons()
        self.inter_hist_plot.showGrid(x=True, y=True)
        self.inter_hist_plot.setLabel("bottom", "Time (s)")
        self.inter_hist_plot.setLabel("left", "Score")
        self.inter_hist_plot.setXRange(-self.history_length_s, 0)
        self.inter_history_smooth_max = et.utils.SmoothMax(self.estimated_frame_rate)
        self.inter_hist_plot.setYRange(0, 10)
        if not self.ref_app_config.inter_enable:
            inter_color = et.utils.color_cycler(0)
            inter_color = f"{inter_color}50"
            inter_dashed_pen = pg.mkPen(inter_color, width=2.5, style=QtCore.Qt.DashLine)
            inter_pen = pg.mkPen(inter_color, width=2)
        else:
            inter_pen = et.utils.pg_pen_cycler(0)
            inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

        self.inter_hist_curve = self.inter_hist_plot.plot(pen=inter_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=inter_dashed_pen)
        self.inter_hist_plot.addItem(limit_line)
        self.inter_limit_lines.append(limit_line)

        for line in self.inter_limit_lines:
            line.setPos(self.ref_app_config.inter_detection_threshold)

        # Sector plot

        self.sector_plot = pg.PlotItem(
            title="Detection zone<br>Detection type: fast (orange), slow (blue), both (green)"
        )
        self.sector_plot.setAspectLocked()
        self.sector_plot.hideAxis("left")
        self.sector_plot.hideAxis("bottom")
        self.sectors = []
        self.limit_text = []

        self.range_html = (
            '<div style="text-align: center">'
            '<span style="color: #000000;font-size:12pt;">'
            "{}</span></div>"
        )

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(self.num_sectors) + 1):
            sector = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            self.sector_plot.addItem(sector)
            self.sectors.append(sector)

            limit = pg.TextItem(html=self.range_html, anchor=(0.5, 0.5), angle=25)
            x = r * np.cos(np.radians(span_deg))
            y = r * np.sin(np.radians(span_deg))
            limit.setPos(x, y + 0.25)
            self.sector_plot.addItem(limit)
            self.limit_text.append(limit)

        self.sectors.reverse()

        start_limit_text = pg.TextItem(html=self.range_html, anchor=(0.5, 0.5), angle=25)
        range_html = self.range_html.format(f"{self.ref_app_config.start_m}")
        start_limit_text.setHtml(range_html)
        start_limit_text.setPos(0, 0.25)
        self.sector_plot.addItem(start_limit_text)

        unit_text = pg.TextItem(html=self.range_html, anchor=(0.5, 0.5))
        unit_html = self.range_html.format("[m]")
        unit_text.setHtml(unit_html)
        unit_text.setPos(
            self.num_sectors + 0.5, (self.num_sectors + 1) * np.sin(np.radians(span_deg))
        )
        self.sector_plot.addItem(unit_text)

        for (text_item, limit) in zip(self.limit_text, np.flip(self.zone_limits)):
            range_html = self.range_html.format(np.around(limit, 1))
            text_item.setHtml(range_html)

        sublayout = win.addLayout(row=1, col=0, colspan=2)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.sector_plot, row=0, col=0)

        self.setup_is_done = True

    def update(self, data):
        # Intra presence

        move_hist_xs = np.linspace(-self.history_length_s, 0, self.history_length_n)

        self.intra_history = np.roll(self.intra_history, -1)
        self.intra_history[-1] = data.intra_presence_score

        m_hist = max(
            float(np.max(self.intra_history)), self.ref_app_config.intra_detection_threshold * 1.05
        )
        m_hist = self.intra_history_smooth_max.update(m_hist)

        self.intra_hist_plot.setYRange(0, m_hist)
        self.intra_hist_curve.setData(move_hist_xs, self.intra_history)

        # Inter presence

        self.inter_history = np.roll(self.inter_history, -1)
        self.inter_history[-1] = data.inter_presence_score

        m_hist = max(
            float(np.max(self.inter_history)), self.ref_app_config.inter_detection_threshold * 1.05
        )
        m_hist = self.inter_history_smooth_max.update(m_hist)

        self.inter_hist_plot.setYRange(0, m_hist)
        self.inter_hist_curve.setData(move_hist_xs, self.inter_history)

        # Sector

        brush = et.utils.pg_brush_cycler(7)
        for sector in self.sectors:
            sector.setBrush(brush)

        if data.presence_detected:
            assert data.max_presence_zone is not None
            if data.max_presence_zone == data.max_intra_zone:
                self.sectors[data.max_presence_zone].setBrush(et.utils.pg_brush_cycler(1))
            else:
                self.sectors[data.max_presence_zone].setBrush(et.utils.pg_brush_cycler(0))


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved
from __future__ import annotations

import typing as t

import h5py
import numpy as np
import numpy.typing as npt

from acconeer.exptool.a121._core import INT_16_COMPLEX, complex_array_to_int16_complex
from acconeer.exptool.a121._core.entities.containers.utils import int16_complex_array_to_complex
from acconeer.exptool.utils import PhonySeries  # type: ignore[import]

from ._detector import DetectorResult
from ._processors import ProcessorResult


_ALL_PROCESSOR_RESULT_FIELDS = (
    "estimated_distances",
    "estimated_strengths",
    "near_edge_status",
    "recorded_threshold_mean_sweep",
    "recorded_threshold_noise_std",
    "direct_leakage",
    "phase_jitter_comp_reference",
    # "extra_result" ignored by default
)

_ALL_DETECTOR_RESULT_FIELDS = (
    "strengths",
    "distances",
    "near_edge_status",
    "processor_results",
    "service_extended_result",
)

_INT_16_COMPLEX_SENTINEL = -(2**15)


def _stack_optional_arraylike(
    sequence: t.Sequence[t.Optional[t.Union[t.List[t.Any], npt.NDArray[t.Any]]]],
    dtype: t.Union[t.Type[float], t.Type[complex]] = float,
) -> npt.NDArray[t.Any]:
    """
    Tries to create an NDArray from a sequence of Optional ArrayLikes, i.e.
    a sequence that looks something like

        [[1 2],
         None,
         [1 2 3 4],
         []
         ...
        ]

    ``None``/``[]`` are replaced with NaN-arrays that has the same dimensions as the
    largest non-NaN array in the sequence:

        [[1   2   NaN Nan],
         [NaN NaN NaN NaN],
         [1   2   3   4],
         [NaN NaN NaN NaN]
         ...
        ]
    """
    SENTINELS = {float: np.NAN, complex: np.NAN + 1j * np.NAN}
    dims = {np.ndim(x) for x in sequence if x is not None}

    if len(dims) > 1:
        raise ValueError(f"arrays in {sequence} are of different dimensions")

    (dim,) = dims
    if dim > 1:
        raise ValueError("cannot handle arrays with ndim > 1")

    lengths = {len(x) for x in sequence if x is not None}
    length = max(lengths)

    data_type = type(sequence[0][0]) if sequence[0] is not None and len(sequence[0]) > 0 else None
    data_type = float if data_type == np.int64 else data_type

    return np.stack(
        [
            np.full((length,), fill_value=SENTINELS[dtype], dtype=dtype)
            if x is None or np.size(x) < 0
            else np.pad(
                np.array(x, dtype=data_type),
                (0, length - len(x)),
                "constant",
                constant_values=(np.NaN, np.NaN),
            )
            for x in sequence
        ]
    )


class ProcessorResultListH5Serializer:
    """
    Reads or writes a distance ProcessorResult from/to a given h5py.Group
    """

    def __init__(
        self,
        group: h5py.Group,
        fields: t.Sequence[str] = _ALL_PROCESSOR_RESULT_FIELDS,
        allow_missing_fields: bool = False,
    ) -> None:
        """
        :param destination: The H5 group that will have datasets read/written to it
        :param fields: The fields to serialize. Has no effect on de-serialization
        :param allow_missing_fields:
            Whether it's acceptable to break type safety during de-serialization
            by replacing missing fields by ``None``. Has no effect on serialization.
        """
        self.group = group
        self.fields = fields
        self.allow_missing_fields = allow_missing_fields

    def serialize(self, results: t.List[ProcessorResult]) -> None:
        if "extra_result" in self.fields:
            raise NotImplementedError(
                "'extra_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        if "estimated_distances" in self.fields:
            data = _stack_optional_arraylike([r.estimated_distances for r in results])

            self.group.create_dataset(
                "estimated_distances",
                dtype=float,
                data=data,
                track_times=False,
            )

        if "estimated_strengths" in self.fields:
            data = _stack_optional_arraylike([r.estimated_strengths for r in results])

            self.group.create_dataset(
                "estimated_strengths",
                dtype=float,
                data=data,
                track_times=False,
            )

        if "near_edge_status" in self.fields:
            near_data = np.array([r.near_edge_status for r in results])

            self.group.create_dataset(
                "near_edge_status",
                dtype=bool,
                data=near_data,
                track_times=False,
            )

        if "recorded_threshold_mean_sweep" in self.fields:
            data = _stack_optional_arraylike([r.recorded_threshold_mean_sweep for r in results])

            self.group.create_dataset(
                "recorded_threshold_mean_sweep",
                dtype=float,
                data=data,
                track_times=False,
            )

        if "recorded_threshold_noise_std" in self.fields:
            data = _stack_optional_arraylike([r.recorded_threshold_noise_std for r in results])

            self.group.create_dataset(
                "recorded_threshold_noise_std",
                dtype=float,
                data=data,
                track_times=False,
            )

        if "direct_leakage" in self.fields:
            data = _stack_optional_arraylike([r.direct_leakage for r in results], dtype=complex)
            # replaces NaNs with int-sentinels before int_16_complex-conversion
            # since NaN becomes 0 at conversion
            data = np.array(
                [
                    np.full(x.shape, fill_value=_INT_16_COMPLEX_SENTINEL * (1 + 1j))
                    if np.isnan(x).all()
                    else x
                    for x in data
                ]
            )

            self.group.create_dataset(
                "direct_leakage",
                dtype=INT_16_COMPLEX,
                data=complex_array_to_int16_complex(data),
                track_times=False,
            )

        if "phase_jitter_comp_reference" in self.fields:
            data = _stack_optional_arraylike(
                [
                    res.phase_jitter_comp_reference.reshape((1, -1)).squeeze()
                    if res.phase_jitter_comp_reference is not None
                    else None
                    for res in results
                ]
            )

            self.group.create_dataset(
                "phase_jitter_comp_reference",
                dtype=float,
                data=data,
                track_times=False,
            )

    @staticmethod
    def _replace_if_all_is_nan(
        x: t.Optional[npt.ArrayLike], replacement: t.Any = None
    ) -> t.Union[npt.ArrayLike, t.Any]:
        if x is None:
            return replacement

        return replacement if np.isnan(x).all() else x

    @staticmethod
    def _direct_leakage_deserialize(
        x: t.Optional[npt.NDArray[t.Any]],
    ) -> t.Optional[npt.NDArray[t.Any]]:
        if x is None:
            return None
        elif (x == np.array(-(2**15), dtype=INT_16_COMPLEX)).all():
            return None
        else:
            return int16_complex_array_to_complex(x)

    def deserialize(self, _: None) -> t.List[ProcessorResult]:
        if "extra_result" in self.fields:
            raise NotImplementedError(
                "'extra_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        groups = (
            self.group.get("estimated_distances", PhonySeries([], is_prototype_singleton=False)),
            self.group.get("estimated_strengths", PhonySeries([], is_prototype_singleton=False)),
            self.group.get("near_edge_status", PhonySeries(None)),
            self.group.get("recorded_threshold_mean_sweep", PhonySeries(None)),
            self.group.get("recorded_threshold_noise_std", PhonySeries(None)),
            self.group.get("direct_leakage", PhonySeries(None)),
            self.group.get("phase_jitter_comp_reference", PhonySeries(None)),
        )

        if any(isinstance(g, PhonySeries) for g in groups) and not self.allow_missing_fields:
            raise ValueError("Some fields are missing while 'allow_missing_fields' is False.")

        if all(isinstance(group, PhonySeries) for group in groups):
            return []

        return [
            ProcessorResult(
                estimated_distances=self._replace_if_all_is_nan(  # type: ignore[arg-type]
                    est_dists,
                ),
                estimated_strengths=self._replace_if_all_is_nan(  # type: ignore[arg-type]
                    est_strengths,
                ),
                near_edge_status=(
                    self._replace_if_all_is_nan(near_edge_status)  # type: ignore[arg-type]
                ),
                recorded_threshold_mean_sweep=(
                    self._replace_if_all_is_nan(rt_mean_sweep)  # type: ignore[arg-type]
                ),
                recorded_threshold_noise_std=(
                    self._replace_if_all_is_nan(rt_noise_std)  # type: ignore[arg-type]
                ),
                direct_leakage=(self._direct_leakage_deserialize(direct_leakage)),
                phase_jitter_comp_reference=(
                    self._replace_if_all_is_nan(  # type: ignore[arg-type]
                        jitter_comp.reshape((-1, 1)) if jitter_comp is not None else None
                    )
                ),
                extra_result=None,  # type: ignore[arg-type]
            )
            for (
                est_dists,
                est_strengths,
                near_edge_status,
                rt_mean_sweep,
                rt_noise_std,
                direct_leakage,
                jitter_comp,
            ) in zip(*groups)
        ]


class DetectorResultListH5Serializer:
    """
    Reads or writes a list of dictionaries of int,
    distance DetectorResult from/to a given h5py.Group
    """

    def __init__(
        self,
        group: h5py.Group,
        fields: t.Sequence[str] = _ALL_DETECTOR_RESULT_FIELDS,
        allow_missing_fields: bool = False,
    ) -> None:
        """
        :param destination: The H5 group that will have datasets read/written to it
        :param fields: The fields to serialize. Has no effect on de-serialization
        :param allow_missing_fields:
            Whether it's acceptable to break type safety during de-serialization
            by replacing missing fields by ``None``. Has no effect on serialization.
        """
        self.group = group
        self.fields = fields
        self.allow_missing_fields = allow_missing_fields

    def serialize(self, results: t.List[t.Dict[int, DetectorResult]]) -> None:
        if "processor_results" in self.fields:
            raise NotImplementedError(
                "'processor_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        if "service_extended_result" in self.fields:
            raise NotImplementedError(
                "'service_extended_result' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        if "strengths" in self.fields:
            data: t.Dict[int, t.List[t.Optional[npt.NDArray[t.Any]]]] = {}
            for res in results:
                for sensor_id, detector_result in res.items():
                    if sensor_id not in data:
                        data[sensor_id] = []
                    data[sensor_id].append(detector_result.strengths)

            strengths_data: t.Dict[int, npt.ArrayLike] = {}
            for sensor_id, strengths in data.items():
                strengths_data[sensor_id] = _stack_optional_arraylike(strengths)

            strengths_group = self.group.create_group("strengths")

            for sens, strengths_item in strengths_data.items():
                strengths_group.create_dataset(
                    str(sens),
                    dtype=float,
                    data=strengths_item,
                    track_times=False,
                )

        if "distances" in self.fields:
            data = {}
            for res in results:
                for sensor_id, detector_result in res.items():
                    if sensor_id not in data:
                        data[sensor_id] = []
                    data[sensor_id].append(detector_result.distances)

            distance_data: t.Dict[int, npt.ArrayLike] = {}
            for sensor_id, distances in data.items():
                distance_data[sensor_id] = _stack_optional_arraylike(distances)

            distances_group = self.group.create_group("distances")

            for sens, dist in distance_data.items():
                distances_group.create_dataset(
                    str(sens),
                    dtype=float,
                    data=dist,
                    track_times=False,
                )

        if "near_edge_status" in self.fields:
            near_data: t.Dict[int, t.List[t.Optional[bool]]] = {}
            for res in results:
                for sensor_id, detector_result in res.items():
                    if sensor_id not in near_data:
                        near_data[sensor_id] = []
                    near_data[sensor_id].append(detector_result.near_edge_status)

            near_edge_data: t.Dict[int, npt.ArrayLike] = {}
            for sensor_id, near_edge in near_data.items():
                near_edge_data[sensor_id] = np.array(near_edge)

            near_edge_group = self.group.create_group("near_edge_status")

            for sens, near in near_edge_data.items():
                near_edge_group.create_dataset(
                    str(sens),
                    dtype=bool,
                    data=near,
                    track_times=False,
                )

    def deserialize(self, _: None) -> t.List[t.Dict[int, DetectorResult]]:
        if "processor_results" in self.fields:
            raise NotImplementedError(
                "'processor_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        if "service_extended_result" in self.fields:
            raise NotImplementedError(
                "'service_extended_result' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        groups = (
            self.group.get("strengths", PhonySeries([], is_prototype_singleton=False)),
            self.group.get("distances", PhonySeries([], is_prototype_singleton=False)),
            self.group.get("near_edge_status", PhonySeries(None)),
        )

        if any(isinstance(g, PhonySeries) for g in groups) and not self.allow_missing_fields:
            raise ValueError("Some fields are missing while 'allow_missing_fields' is False.")

        if all(isinstance(group, PhonySeries) for group in groups):
            return []

        res = []
        (strengths, dists, near_edge_status) = groups

        for k in strengths.keys():
            for r, d, n in zip(strengths.get(k), dists.get(k), near_edge_status.get(k)):
                res.append(
                    {
                        int(k): DetectorResult(
                            strengths=r[~np.isnan(r)],
                            distances=d[~np.isnan(d)],
                            near_edge_status=(n),
                            processor_results=[],
                            service_extended_result=[],
                        )
                    }
                )

        return res



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import importlib
import logging
from typing import Any, Optional, Tuple

from requests import Response, Session
from requests.cookies import RequestsCookieJar

from PySide6.QtCore import Qt
from PySide6.QtGui import QCloseEvent, QMovie
from PySide6.QtWidgets import (
    QComboBox,
    QFileDialog,
    QGridLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app import resources  # type: ignore[attr-defined]
from acconeer.exptool.app.new._enums import ConnectionInterface
from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.ui import utils as ui_utils
from acconeer.exptool.app.new.ui.device_comboboxes import SerialPortComboBox, USBDeviceComboBox
from acconeer.exptool.app.new.ui.plugin_components.utils import GroupBox
from acconeer.exptool.flash import (  # type: ignore[import]
    DevLicense,
    clear_cookies,
    get_flash_download_name,
    get_flash_known_devices,
)
from acconeer.exptool.flash._products import (  # type: ignore[import]
    EVK_TO_PRODUCT_MAP,
    PRODUCT_NAME_TO_FLASH_MAP,
)
from acconeer.exptool.utils import CommDevice, SerialDevice  # type: ignore[import]

from .dialogs import FlashDialog, FlashLoginDialog, LicenseAgreementDialog, UserMessageDialog
from .threads import AuthThread, BinDownloadThread


_LOGGED_IN_MSG = (
    "Logged in to <a href='https://developer.acconeer.com/'>developer.acconeer.com</a>"
)
log = logging.getLogger(__name__)


class FlashMainWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground, True)
        self.setObjectName(type(self).__name__)
        self.setStyleSheet(f"*#{type(self).__name__} " + "{ background-color: #11515766; }")

        self.app_model = app_model

        self.bin_file: Optional[str] = None

        self.authenticating = False
        self.downloading_firmware = False
        self.dev_license = DevLicense()

        browse_button = QPushButton("Browse", self)
        browse_button.clicked.connect(self._browse_file)

        self.file_label = QLineEdit(self)
        self.file_label.setReadOnly(True)
        self.file_label.setEnabled(False)
        self.file_label.setPlaceholderText("<Select a bin file>")

        self.get_latest_button = QPushButton("Get latest bin file", self)
        self.get_latest_button.clicked.connect(self._get_latest_bin_file)

        self.downloaded_version_label = QLineEdit(self)
        self.downloaded_version_label.setReadOnly(True)
        self._reset_download_version()

        self.download_spinner = QLabel()
        self.download_spinner.setHidden(True)

        loader_gif = None
        with importlib.resources.path(resources, "loader.gif") as path:
            loader_gif = path

        self.download_movie = QMovie(str(loader_gif))
        self.download_spinner.setMovie(self.download_movie)

        self.download_status_label = QLabel()
        self.download_status_label.setHidden(True)

        self.interface_dd = QComboBox(self)
        self.interface_dd.addItem("Serial", userData=ConnectionInterface.SERIAL)
        self.interface_dd.addItem("USB", userData=ConnectionInterface.USB)

        self.interface_dd.currentIndexChanged.connect(self._on_interface_dd_change)

        self.device_selection = QComboBox(self)
        self.device_selection.setEnabled(False)
        self.device_name_selection = None
        self.device_selection.currentIndexChanged.connect(self._on_device_selection_change)

        self.stacked = QStackedWidget(self)
        self.stacked.setStyleSheet("QStackedWidget {background-color: transparent;}")
        self.stacked.addWidget(SerialPortComboBox(app_model, self.stacked))
        self.stacked.addWidget(USBDeviceComboBox(app_model, self.stacked))

        self.flash_button = QPushButton("Flash", self)
        self.flash_button.clicked.connect(self._flash)
        self.flash_button.setEnabled(False)

        self.logged_in_label = QLabel(self)
        self.logged_in_label.setOpenExternalLinks(True)

        self.logout_button = QPushButton("Logout", self)
        self.logout_button.clicked.connect(self._logout)
        self._hide_login_status()

        button_layout = QGridLayout()

        # fmt: off
        # Grid layout:                                  row, col, rspan, cspan
        button_layout.addWidget(browse_button,                 0,   0,   1,     3)    # noqa: E241
        button_layout.addWidget(self.file_label,               0,   3,   1,     9)    # noqa: E241
        button_layout.addWidget(self.get_latest_button,        1,   0,   1,     3)    # noqa: E241
        button_layout.addWidget(self.downloaded_version_label, 1,   3,   1,     9)    # noqa: E241
        button_layout.addWidget(self.interface_dd,             2,   0,   1,     3)    # noqa: E241
        button_layout.addWidget(self.stacked,                  2,   3,   1,     6)    # noqa: E241
        button_layout.addWidget(self.device_selection,         2,   9,   1,     3)    # noqa: E241
        button_layout.addWidget(self.flash_button,             3,   0,   1,    12)    # noqa: E241
        button_layout.addWidget(self.logged_in_label,          4,   0,   1,     8)    # noqa: E241
        button_layout.addWidget(self.logout_button,            4,   8,   1,     4)    # noqa: E241
        button_layout.addWidget(self.download_spinner,         5,   0,   1,     1)    # noqa: E241
        button_layout.addWidget(self.download_status_label,    5,   1,   1,    11)    # noqa: E241
        # fmt: on

        box = GroupBox.vertical("", parent=self)
        box.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        box.layout().addWidget(ui_utils.LayoutWrapper(button_layout))

        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(box)

        self.setLayout(layout)

        self.browse_file_dialog = QFileDialog(None)
        self.browse_file_dialog.setNameFilter("Bin files (*.bin)")

        self.flash_dialog = FlashDialog(self)
        self.flash_dialog.flash_done.connect(self._flash_done)

        app_model.sig_notify.connect(self._on_app_model_update)

    @property
    def flash_device(self) -> Optional[CommDevice]:
        if self.app_model.connection_interface == ConnectionInterface.SERIAL:
            return self.app_model.serial_connection_device
        if self.app_model.connection_interface == ConnectionInterface.USB:
            return self.app_model.usb_connection_device
        return None

    @property
    def device_name(self) -> Optional[str]:
        if not self.device_selection.isEnabled():
            return None
        else:
            return self.device_name_selection

    def _get_latest_bin_file(self) -> None:
        self.auth_thread = AuthThread(dev_license=self.dev_license)
        self.auth_thread.started.connect(self._auth_start)
        self.auth_thread.finished.connect(self.auth_thread.deleteLater)
        self.auth_thread.finished.connect(self._auth_stop)
        self.auth_thread.license_loaded.connect(self._license_loaded)
        self.auth_thread.auth_done.connect(self._auth_done)
        self.auth_thread.auth_failed.connect(self._auth_failed)

        self.authenticating = True
        self.auth_thread.start()

    def _auth_start(self) -> None:
        self._show_download_progress("Authenticating...")

    def _auth_stop(self) -> None:
        self._hide_download_progress()
        self.authenticating = False

    def _license_loaded(self, dev_license: DevLicense) -> None:
        self.dev_license = dev_license

    def _auth_done(
        self, auth_info: Tuple[RequestsCookieJar, Tuple[bool, Session, Response]]
    ) -> None:
        self._show_login_status(_LOGGED_IN_MSG)

        cookies, content = auth_info
        self._init_download(cookies, content)

    def _auth_failed(self) -> None:
        login_dialog = FlashLoginDialog(self)
        authenticated = login_dialog.exec()

        if authenticated:
            self._show_login_status(_LOGGED_IN_MSG)
            cookies, content = login_dialog.get_auth_info()
            if cookies is not None and content is not None:
                self._init_download(cookies, content)

    def _init_download(
        self, cookies: RequestsCookieJar, content: Tuple[bool, Session, Response]
    ) -> None:

        license_agreement_dialog = LicenseAgreementDialog(self.dev_license, self)
        license_accepted = license_agreement_dialog.exec()
        device_download_name = get_flash_download_name(self.flash_device, self.device_name)

        if license_accepted:
            self.download_thread = BinDownloadThread(
                device_download_name, cookies, content[1], content[2]
            )

            self.download_thread.started.connect(self._download_start)
            self.download_thread.finished.connect(self.download_thread.deleteLater)
            self.download_thread.finished.connect(self._download_stop)
            self.download_thread.download_done.connect(self._download_done)
            self.download_thread.download_failed.connect(self._download_failed)

            self.downloading_firmware = True
            self.download_thread.start()

    def _download_start(self) -> None:
        self._show_download_progress("Downloading image file...")

    def _download_stop(self) -> None:
        self._hide_download_progress()
        self.downloading_firmware = False

    def _download_done(self, bin_file: str, version: str) -> None:
        self.bin_file = bin_file
        self._set_version(version)
        self.downloading_firmware = False
        self._draw()

    def _download_failed(self, error_msg: str) -> None:
        self.bin_file = None
        log.error(f"Failed to download firmware: {error_msg}")
        self._draw()

    def _browse_file(self) -> None:
        if self.browse_file_dialog.exec():
            filenames = self.browse_file_dialog.selectedFiles()
            self.bin_file = filenames[0]
            self._reset_download_version()

        self._draw()

    def _show_download_progress(self, text: str) -> None:
        self.download_movie.start()
        self.download_spinner.setHidden(False)
        self.download_status_label.setText(text)
        self.download_status_label.setHidden(False)

    def _hide_download_progress(self) -> None:
        self.download_movie.stop()
        self.download_spinner.setHidden(True)
        self.download_status_label.setHidden(True)
        self.adjustSize()

    def _flash(self) -> None:
        assert self.bin_file is not None

        boot_description = self._get_boot_description(self.flash_device, self.device_name)
        if boot_description:
            UserMessageDialog(
                "Bootloader description",
                boot_description,
                "Got it! The board is in bootloader mode",
                self,
            ).exec()

        self.app_model.set_port_updates_pause(True)
        self.flash_dialog.flash(self.bin_file, self.flash_device, self.device_name)

    def _flash_done(self) -> None:
        self.app_model.set_port_updates_pause(False)

    def _on_interface_dd_change(self) -> None:
        self.app_model.set_connection_interface(self.interface_dd.currentData())

    def _on_device_selection_change(self) -> None:
        self.device_name_selection = self.device_selection.currentText()

    def _on_app_model_update(self, app_model: AppModel) -> None:
        if app_model.connection_interface in [ConnectionInterface.SERIAL, ConnectionInterface.USB]:
            interface_index = self.interface_dd.findData(app_model.connection_interface)
            if interface_index == -1:
                raise RuntimeError

            self.interface_dd.setCurrentIndex(interface_index)
            self.stacked.setCurrentIndex(interface_index)

            current_index = self.device_selection.currentIndex()
            if self.flash_device is None or (
                self.flash_device is not None and self.flash_device.name is not None
            ):
                self.device_selection.clear()
                self.device_selection.setEnabled(False)
            else:
                self.device_selection.clear()
                for dev in get_flash_known_devices():
                    self.device_selection.addItem(dev)
                self.device_selection.setEnabled(True)
                if current_index >= 0:
                    self.device_selection.setCurrentIndex(current_index)

            enable_select = not self.authenticating and not self.downloading_firmware
            self.get_latest_button.setEnabled(enable_select)
        else:
            self.device_selection.clear()
            self.device_selection.setEnabled(False)

        self._draw()

    def _draw(self) -> None:
        self.file_label.setText(self.bin_file if self.bin_file else "")
        self.file_label.setEnabled(self.bin_file is not None)
        self.flash_button.setEnabled(
            not self.authenticating
            and not self.downloading_firmware
            and self.flash_device is not None
            and self.bin_file is not None
        )

    def exec(self) -> None:
        if self.app_model.connection_interface == ConnectionInterface.SOCKET:
            self.app_model.set_connection_interface(ConnectionInterface.USB)

        super().exec()

    def closeEvent(self, event: QCloseEvent) -> None:
        if self.authenticating:
            self.auth_thread.terminate()
            self.auth_thread.wait()

        if self.downloading_firmware:
            self.download_thread.terminate()
            self.download_thread.wait()
        super().closeEvent(event)

    def _show_login_status(self, msg: str, disable_logout: bool = False) -> None:
        self.logged_in_label.setText(msg)
        self.logged_in_label.setHidden(False)
        self.logout_button.setHidden(False)
        self.logout_button.setEnabled(not disable_logout)

    def _hide_login_status(self) -> None:
        self.logged_in_label.setHidden(True)
        self.logout_button.setHidden(True)

    def _logout(self) -> None:
        clear_cookies()
        self._show_login_status("Currently not logged in", disable_logout=True)

    def _set_version(self, version: str) -> None:
        self.downloaded_version_label.setText(version)
        self.downloaded_version_label.setEnabled(True)

    def _reset_download_version(self) -> None:
        self.downloaded_version_label.setEnabled(False)
        self.downloaded_version_label.setText("<Downloaded bin file version>")

    def _get_boot_description(self, flash_device: SerialDevice, device_name: Optional[str]) -> Any:

        flash_device_name: str = device_name or flash_device.name
        product: Optional[str] = None

        if flash_device_name in EVK_TO_PRODUCT_MAP.keys():
            product = EVK_TO_PRODUCT_MAP[flash_device_name]

        if product in PRODUCT_NAME_TO_FLASH_MAP.keys():
            return PRODUCT_NAME_TO_FLASH_MAP[product].get_boot_description(product)

        return None



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

from typing import Any, Callable, Optional, TypeVar, Union, cast

from acconeer.exptool.a121 import Client, Result, _RateCalculator
from acconeer.exptool.a121._core import utils

from ._message import GeneralMessage, Message


ClientT = TypeVar("ClientT", bound=Client)


class ApplicationClient:

    callback: Callable[[Message], None]
    _wrapped_client: Client
    _rate_stats_calc: Optional[_RateCalculator]
    _frame_count: int

    def __init__(self, wrapped_client: Client, callback: Callable[[Message], None]) -> None:
        self._wrapped_client = wrapped_client
        self.callback = callback
        self._rate_stats_calc = None
        self._frame_count = 0

    def __getattr__(self, name: str) -> Any:
        """
        This is the key to the "fall-through-wrapping".
        """
        return getattr(self._wrapped_client, name)

    @classmethod
    def wrap(cls, wrapped_client: ClientT, callback: Callable[[Message], None]) -> ClientT:
        """
        Factory that makes the ApplicationClient transparent from a
        typing perspective.

        The cast is necessary as we are dealing with magic here, but should hold
        up so long any overriden functions have the same type as the overriders.
        """
        return cast(ClientT, cls(wrapped_client, callback))

    def start_session(self) -> None:
        self._wrapped_client.start_session()
        assert self._wrapped_client.session_config is not None
        assert self._wrapped_client.extended_metadata is not None

        if self._wrapped_client.session_config.extended:
            self._rate_stats_calc = _RateCalculator(
                self._wrapped_client.session_config, self._wrapped_client.extended_metadata
            )
        else:
            metadata = utils.unextend(self._wrapped_client.extended_metadata)
            self._rate_stats_calc = _RateCalculator(self._wrapped_client.session_config, metadata)

    def get_next(self) -> Union[Result, list[dict[int, Result]]]:
        results = self._wrapped_client.get_next()
        assert self._rate_stats_calc is not None
        self._rate_stats_calc.update(results)
        self.callback(GeneralMessage(name="rate_stats", data=self._rate_stats_calc.stats))
        self._frame_count += 1
        self.callback(GeneralMessage(name="frame_count", data=self._frame_count))
        return results

    def stop_session(self) -> None:
        result = self._wrapped_client.stop_session()
        self._frame_count = 0
        self.callback(GeneralMessage(name="rate_stats", data=None))
        self.callback(GeneralMessage(name="frame_count", data=None))
        return result



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

import acconeer.exptool as et
from acconeer.exptool import a121


parser = a121.ExampleArgumentParser()
parser.add_argument("--output-file", required=True)
parser.add_argument("--num-frames", required=True, type=int)
args = parser.parse_args()
et.utils.config_logging(args)

client = a121.Client.open(**a121.get_client_args(args))

# Session setup, just like the other examples.
config = a121.SessionConfig()
client.setup_session(config)

# Here we create and attach a H5Recorder to the Client.
# The H5Recorder will sample all frames retrieved in the client.
# Once the "with"-block have been exited, the H5Recorder will
# wrap up and close the file.
with a121.H5Recorder(args.output_file, client):
    client.start_session()

    for i in range(args.num_frames):
        # Client will send its Results to the H5Recorder.
        client.get_next()
        print(f"Result {i + 1}/{args.num_frames} was sampled")

    client.stop_session()

client.close()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import copy
import logging
from typing import Any, Optional

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool.a121._core import Criticality

from . import pidgets
from .data_editor import DataEditor
from .json_save_load_buttons import JsonSaveLoadButtons
from .sensor_config_editor import SensorConfigEditor
from .utils import GroupBox


log = logging.getLogger(__name__)


class SessionConfigEditor(DataEditor[Optional[a121.SessionConfig]]):
    _session_config: Optional[a121.SessionConfig]
    _server_info: Optional[a121.ServerInfo]
    _sensor_id_pidget: pidgets.SensorIdPidget

    _update_rate_erroneous: bool

    sig_update = Signal(object)

    SPACING = 15

    def __init__(
        self, supports_multiple_subsweeps: bool = False, parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(parent=parent)

        self._server_info = None

        self._session_config = None
        self._update_rate_erroneous = False

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.session_group_box = GroupBox.vertical(
            "Session parameters",
            JsonSaveLoadButtons.from_editor_and_config_type(self, a121.SessionConfig),
            parent=self,
        )
        self.session_group_box.layout().setSpacing(self.SPACING)
        self.layout().addWidget(self.session_group_box)

        self._sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(self)
        self._sensor_id_pidget.sig_update.connect(self._update_sole_sensor_id)
        self.session_group_box.layout().addWidget(self._sensor_id_pidget)

        self._update_rate_pidget = pidgets.OptionalFloatPidgetFactory(
            name_label_text="Update rate:",
            name_label_tooltip=(
                "Set an update rate limit on the server.\n"
                "If 'Limit' is unchecked, the server will run as fast as possible."
            ),
            limits=(0.1, 1e4),
            decimals=1,
            init_set_value=10.0,
            placeholder_text="- Hz",
            suffix="Hz",
            checkbox_label_text="Limit",
        ).create(self)
        self._update_rate_pidget.sig_update.connect(self._update_update_rate)
        self.session_group_box.layout().addWidget(self._update_rate_pidget)

        self._sensor_config_editor = SensorConfigEditor(supports_multiple_subsweeps, self)
        self._sensor_config_editor.sig_update.connect(self._update_sole_sensor_config)
        self.layout().addWidget(self._sensor_config_editor)

    def set_selected_sensor(self, sensor_id: Optional[int], sensor_list: list[int]) -> None:
        self._sensor_id_pidget.set_selected_sensor(sensor_id, sensor_list)

    def set_read_only(self, read_only: bool) -> None:
        self._sensor_id_pidget.setEnabled(not read_only)
        self._update_rate_pidget.setEnabled(not read_only)
        self._sensor_config_editor.set_read_only(read_only)

    def set_data(self, session_config: Optional[a121.SessionConfig]) -> None:
        if self._session_config == session_config:
            return

        self._session_config = session_config
        self.setEnabled(session_config is not None)

        if self._session_config is None:
            log.debug("could not update ui as SessionConfig is None")
            return

        self._update_rate_pidget.set_data(self._session_config.update_rate)
        self._sensor_config_editor.set_data(self._session_config.sensor_config)

    def get_data(self) -> Optional[a121.SessionConfig]:
        return copy.deepcopy(self._session_config)

    def handle_validation_results(
        self, results: list[a121.ValidationResult]
    ) -> list[a121.ValidationResult]:
        self._update_rate_pidget.set_note_text(None)
        self._sensor_id_pidget.set_note_text(None)

        unhandled_results: list[a121.ValidationResult] = []

        for result in results:
            if not self._handle_validation_result(result):
                unhandled_results.append(result)

        unhandled_results = self._sensor_config_editor.handle_validation_results(unhandled_results)

        return unhandled_results

    @property
    def is_ready(self) -> bool:
        return not self._update_rate_erroneous and self._sensor_config_editor.is_ready

    def _handle_validation_result(self, result: a121.ValidationResult) -> bool:
        if self._session_config is None:
            raise RuntimeError(
                "SessionConfigEditor's config is None while ValidationResults are being handled."
            )

        result_handled = False

        if result.aspect == "update_rate":
            self._update_rate_pidget.set_note_text(result.message, result.criticality)
            result_handled = True
        elif result.aspect == "sensor_id":
            self._sensor_id_pidget.set_note_text(result.message, result.criticality)
            result_handled = True

        return result_handled

    def _update_update_rate(self, value: Any) -> None:
        if self._session_config is None:
            raise TypeError("SessionConfig is None")

        config = copy.deepcopy(self._session_config)

        try:
            config.update_rate = value
        except Exception as e:
            self._update_rate_erroneous = True
            self._update_rate_pidget.set_note_text(e.args[0], Criticality.ERROR)

            # this emit needs to be done to signal that "is_ready" has changed.
            self.sig_update.emit(self.get_data())
        else:
            self._update_rate_erroneous = False

            self.set_data(config)
            self.sig_update.emit(config)

    def _update_sole_sensor_id(self, value: Any) -> None:
        if self._session_config is None:
            raise TypeError("SessionConfig is None")

        config = copy.deepcopy(self._session_config)

        try:
            config.sensor_id = value
        except Exception as e:
            self._sensor_id_pidget.set_note_text(e.args[0], Criticality.ERROR)

        self.set_data(config)
        self.sig_update.emit(config)

    def _update_sole_sensor_config(self, sensor_config: a121.SensorConfig) -> None:
        if self._session_config is None:
            raise RuntimeError

        config = a121.SessionConfig(
            {self._session_config.sensor_id: sensor_config},
            update_rate=self._session_config.update_rate,
            extended=self._session_config.extended,
        )
        self.set_data(config)
        self.sig_update.emit(config)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import enum
import functools
import re
import typing as t

import typing_extensions as te

from PySide6.QtCore import Qt
from PySide6.QtGui import QColor, QFontMetricsF, QSyntaxHighlighter, QTextCharFormat, QTextDocument
from PySide6.QtWidgets import QGridLayout, QLabel, QPushButton, QTextEdit, QWidget


class PresentationType(enum.Enum):
    """
    Enum describing a presentation type.

    A PresenterFunc accept an PresentationType instance as its second argument.
    """

    JSON = "JSON"
    C_SET_CONFIG = "C set_config"

    @property
    def button_label(self) -> str:
        return self.value


PresenterFunc = t.Callable[[t.Any, PresentationType], t.Optional[str]]


class _Rgb(te.TypedDict, total=True):
    r: int
    g: int
    b: int


class _SyntaxHighlighter(QSyntaxHighlighter):
    GOLDEN_ROD: _Rgb = dict(r=218, g=165, b=32)
    PURPLE: _Rgb = dict(r=160, g=32, b=240)
    FOREST_GREEN: _Rgb = dict(r=34, g=139, b=34)
    DODGER_BLUE: _Rgb = dict(r=30, g=144, b=255)
    FIREBRICK: _Rgb = dict(r=178, g=34, b=34)
    SIENNA: _Rgb = dict(r=160, g=82, b=45)
    GRAY: _Rgb = dict(r=100, g=100, b=100)

    def __init__(self, parent: QTextDocument, highlights: dict[str, QTextCharFormat]) -> None:
        super().__init__(parent)
        self._highlights = highlights

    def highlightBlock(self, text: str) -> None:
        for pattern, fmt in self._highlights.items():
            for match in re.finditer(pattern, text):
                # The first capturing group in the regex will be styled
                try:
                    match_length = match.end(1) - match.start(1)
                    self.setFormat(match.start(1), match_length, fmt)
                except IndexError:
                    pass

    @staticmethod
    def _create_char_fmt(
        *,
        r: int = 0,
        g: int = 0,
        b: int = 0,
        italic: bool = False,
        weight: int = 400,
    ) -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(QColor(r, g, b))
        fmt.setFontItalic(italic)
        fmt.setFontWeight(weight)
        return fmt

    @classmethod
    def json(cls, parent: QTextDocument) -> _SyntaxHighlighter:
        return cls(
            parent,
            {
                r'("[^"]*")': cls._create_char_fmt(**cls.GOLDEN_ROD),  # string literals
                r"(\d+(?:\.\d+)?),": cls._create_char_fmt(**cls.DODGER_BLUE),  # number literals
            },
        )

    @classmethod
    def rss_c(cls, parent: QTextDocument) -> _SyntaxHighlighter:
        snake_case = "[a-z]+(?:_[a-z]+)*"
        type_regex = f"{snake_case}_t "
        acc_type_regex = f"acc_{type_regex}"
        return cls(
            parent,
            {
                # anything snake_case
                rf"({snake_case})": cls._create_char_fmt(**cls.SIENNA),
                # keywords
                r"(static)": cls._create_char_fmt(**cls.PURPLE),
                # types
                rf"(void|{type_regex}|{acc_type_regex})": cls._create_char_fmt(**cls.FOREST_GREEN),
                # function call & -definition
                rf"(set_config|acc_{snake_case}_set)\(": cls._create_char_fmt(b=255, weight=500),
                # int literals
                r"(\d+U?)": cls._create_char_fmt(**cls.DODGER_BLUE),
                # floating literals
                r"(\d+\.\d+f?)": cls._create_char_fmt(**cls.DODGER_BLUE),
                # bool literals
                "(true|false)": cls._create_char_fmt(r=0, g=0, b=0),
                # acconeer enum members/constants
                r"(ACC[_A-Z0-9]*)": cls._create_char_fmt(**cls.GOLDEN_ROD),
                # single line comments
                r"(//.*)": cls._create_char_fmt(**cls.FIREBRICK, italic=True),
                # parameter ignores
                rf"(\(void\){snake_case};)": cls._create_char_fmt(**cls.GRAY, italic=True),
            },
        )


class PresentationWindow(QWidget):
    _NOTE_TEXT = "<b>Note!</b> A compact JSON string is always saved to file."
    _HIGHLIGHTERS = {
        PresentationType.JSON: _SyntaxHighlighter.json,
        PresentationType.C_SET_CONFIG: _SyntaxHighlighter.rss_c,
    }

    def __init__(
        self, presentations: dict[PresentationType, str], parent: t.Optional[QWidget] = None
    ) -> None:
        super().__init__(parent=parent)

        self._preview = QTextEdit(parent=self)
        self._preview.setFontFamily("monospace")
        self._preview.setReadOnly(True)
        self._preview.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)
        self._highlighter = _SyntaxHighlighter.json(self._preview.document())

        tab_stop_distance = QFontMetricsF(self._preview.font()).horizontalAdvance(" ") * 8
        self._preview.setTabStopDistance(tab_stop_distance)
        self._preview.setText(
            presentations.get(PresentationType.JSON, "No JSON presentation available")
        )

        layout = QGridLayout()

        num_columns = len(PresentationType)
        layout.addWidget(QLabel(f"<i>{self._NOTE_TEXT}</i>"), 0, 0, 1, num_columns)
        layout.addWidget(self._preview, 1, 0, 1, num_columns)
        for idx, member in enumerate(PresentationType):
            presentation_exists = member in presentations
            button_label = member.button_label + (
                "" if presentation_exists else " (Not available)"
            )

            button = QPushButton(button_label)
            button.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            button.setEnabled(presentation_exists)
            button.clicked.connect(
                functools.partial(self._preview.setText, presentations.get(member, ""))
            )
            button.clicked.connect(functools.partial(self._swap_highlighter, member))
            layout.addWidget(button, 2, idx)

        self.setLayout(layout)

    def _swap_highlighter(self, presentation_type: PresentationType) -> None:
        self._highlighter.deleteLater()
        self._highlighter = self._HIGHLIGHTERS[presentation_type](self._preview.document())



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.bilateration import Processor, ProcessorConfig
from acconeer.exptool.a121.algo.distance import Detector, DetectorConfig, ThresholdMethod


_SENSOR_IDS = [2, 3]


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client.open(**a121.get_client_args(args))

    detector_config = DetectorConfig(
        start_m=0.25,
        end_m=1.0,
        signal_quality=25.0,
        max_profile=a121.Profile.PROFILE_1,
        max_step_length=2,
        threshold_method=ThresholdMethod.CFAR,
    )

    detector = Detector(client=client, sensor_ids=_SENSOR_IDS, detector_config=detector_config)

    session_config = detector.session_config

    processor_config = ProcessorConfig()

    processor = Processor(
        session_config=session_config, processor_config=processor_config, sensor_ids=_SENSOR_IDS
    )

    detector.calibrate_detector()

    detector.start()

    pg_updater = PlotPlugin(detector_config=detector_config, bilateration_config=processor_config)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        detector_result = detector.get_next()
        processor_result = processor.process(detector_result)

        try:
            pg_process.put_data(
                {"detector_result": detector_result, "processor_result": processor_result}
            )
        except et.PGProccessDiedException:
            break

    detector.stop()

    print("Disconnecting...")
    client.close()


class PlotPlugin:

    _NUM_POINTS_ON_CIRCLE = 100

    def __init__(
        self,
        bilateration_config: ProcessorConfig,
        detector_config: DetectorConfig,
    ) -> None:

        self.num_curves = 5
        self.detector_config = detector_config
        self.sensor_half_spacing_m = bilateration_config.sensor_spacing_m / 2

    def setup(self, win):

        # Define sweep plot.
        self.sweep_plot = win.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.setLabel("bottom", "Distance (m)")
        self.sweep_plot.addItem(pg.PlotDataItem())

        pen_sweep_0 = et.utils.pg_pen_cycler(0)
        pen_sweep_1 = et.utils.pg_pen_cycler(1)

        # Add sweep curves for the two sensors.
        feat_kw_1 = dict(pen=pen_sweep_0)
        feat_kw_2 = dict(pen=pen_sweep_1)
        self.sweep_curves_1 = [self.sweep_plot.plot(**feat_kw_1) for _ in range(self.num_curves)]
        self.sweep_curves_2 = [self.sweep_plot.plot(**feat_kw_2) for _ in range(self.num_curves)]

        pen_sweep_0 = et.utils.pg_pen_cycler(0, "--")
        pen_sweep_1 = et.utils.pg_pen_cycler(1, "--")

        # Add threshold curves for the two sensors.
        feat_kw_1 = dict(pen=pen_sweep_0)
        feat_kw_2 = dict(pen=pen_sweep_1)
        self.threshold_curves_1 = [
            self.sweep_plot.plot(**feat_kw_1) for _ in range(self.num_curves)
        ]
        self.threshold_curves_2 = [
            self.sweep_plot.plot(**feat_kw_2) for _ in range(self.num_curves)
        ]

        # Add legends.
        sweep_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        sweep_plot_legend.setParentItem(self.sweep_plot)
        sweep_plot_legend.addItem(self.sweep_curves_1[0], "Sweep - Left sensor")
        sweep_plot_legend.addItem(self.threshold_curves_1[0], "Threshold - Left sensor")
        sweep_plot_legend.addItem(self.sweep_curves_2[0], "Sweep - Right sensor")
        sweep_plot_legend.addItem(self.threshold_curves_2[0], "Threshold - Right sensor")

        self.sweep_smooth_max = et.utils.SmoothMax()

        # Define obstacle plot.
        self.obstacle_location_plot = win.addPlot(row=1, col=0)
        self.obstacle_location_plot.setMenuEnabled(False)
        self.obstacle_location_plot.setAspectLocked()
        self.obstacle_location_plot.showGrid(x=True, y=True)
        self.obstacle_location_plot.addLegend()
        self.obstacle_location_plot.setLabel("left", "Y (m)")
        self.obstacle_location_plot.setLabel("bottom", "X (m)")
        self.obstacle_location_plot.addItem(pg.PlotDataItem())
        self.obstacle_location_plot.setXRange(
            -self.detector_config.end_m, self.detector_config.end_m
        )
        self.obstacle_location_plot.setYRange(0.0, self.detector_config.end_m)

        pen_sweep_0 = et.utils.pg_pen_cycler(0)
        brush_sweep_0 = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush_sweep_0, symbolPen=None)
        feat_kw = dict(pen=None, **symbol_kw)
        self.obstacle_location_curve = self.obstacle_location_plot.plot(**feat_kw)
        feat_kw = dict(pen=pen_sweep_0)
        self.obstacle_location_half_curve = [
            self.obstacle_location_plot.plot(**feat_kw) for _ in range(Processor._MAX_NUM_OBJECTS)
        ]

    def update(self, data) -> None:

        detector_result = data["detector_result"]
        processor_result = data["processor_result"]

        # Plot sweep data from both distance detectors.
        max_val = 0.0
        for distance_detector_result, sweep_curves, threshold_curves in zip(
            detector_result.values(),
            [self.sweep_curves_1, self.sweep_curves_2],
            [self.threshold_curves_1, self.threshold_curves_2],
        ):
            for idx, distance_processor_result in enumerate(
                distance_detector_result.processor_results
            ):

                abs_sweep = distance_processor_result.extra_result.abs_sweep
                threshold = distance_processor_result.extra_result.used_threshold
                distances_m = distance_processor_result.extra_result.distances_m

                assert abs_sweep is not None
                assert threshold is not None
                assert distances_m is not None

                sweep_curves[idx].setData(distances_m, abs_sweep)
                threshold_curves[idx].setData(distances_m, threshold)

                if max_val < np.max(abs_sweep):
                    max_val = float(np.max(abs_sweep))

                if max_val < np.max(threshold):
                    max_val = float(np.max(threshold))

        if max_val != 0.0:
            self.sweep_plot.setYRange(0.0, self.sweep_smooth_max.update(max_val))

        # Plot result from bilateration processor.
        # Start with the points.
        points = processor_result.points
        xs = [point.x_coord for point in points]
        ys = [point.y_coord for point in points]
        self.obstacle_location_curve.setData(xs, ys)

        # Plot objects without counter part.
        objects_without_counterpart = processor_result.objects_without_counterpart
        for i, o in enumerate(objects_without_counterpart):
            x = np.cos(np.linspace(0, np.pi, self._NUM_POINTS_ON_CIRCLE)) * o.distance
            y = np.sin(np.linspace(0, np.pi, self._NUM_POINTS_ON_CIRCLE)) * o.distance

            # Offset circle to center around the sensor that detected the object.
            if o.sensor_position == Processor._SENSOR_POSITION_LEFT:
                x -= self.sensor_half_spacing_m
            elif o.sensor_position == Processor._SENSOR_POSITION_RIGHT:
                x += self.sensor_half_spacing_m
            else:
                raise ValueError("Invalid sensor position.")

            self.obstacle_location_half_curve[i].setData(x, y)

        # Remove curves that does not have an object to visualize.
        for i in range(len(objects_without_counterpart), Processor._MAX_NUM_OBJECTS):
            self.obstacle_location_half_curve[i].setData([], [])


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo import presence, smart_presence


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client.open(**a121.get_client_args(args))

    processor_config = smart_presence.ProcessorConfig(
        num_zones=3,
    )

    detector_config = presence.DetectorConfig(
        start_m=1,
        end_m=3,
    )

    presence_detector = presence.Detector(
        client=client, sensor_id=1, detector_config=detector_config
    )
    presence_detector.start()

    smart_presence_processor = smart_presence.Processor(
        processor_config,
        detector_config,
        presence_detector.session_config,
        presence_detector.detector_metadata,
    )

    pg_updater = PGUpdater(
        detector_config,
        processor_config,
        presence_detector._get_sensor_config(detector_config),
        presence_detector.estimated_frame_rate,
        smart_presence_processor.zone_limits,
    )
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        detector_result = presence_detector.get_next()
        processor_result = smart_presence_processor.process(detector_result)
        if detector_result.presence_detected:
            print(f"Presence in zone {processor_result.max_presence_zone}")
        else:
            print("No presence")

        data = {
            "processor_result": processor_result,
            "detector_result": detector_result,
        }
        try:
            pg_process.put_data(data)
        except et.PGProccessDiedException:
            break

    presence_detector.stop()

    print("Disconnecting...")
    client.close()


class PGUpdater:
    def __init__(
        self,
        detector_config: presence.DetectorConfig,
        processor_config: smart_presence.ProcessorConfig,
        sensor_config: a121.SensorConfig,
        estimated_frame_rate: float,
        zone_limits: npt.NDArray[np.float_],
    ):
        self.detector_config = detector_config

        self.history_length_s = 5
        self.estimated_frame_rate = estimated_frame_rate
        self.history_length_n = int(round(self.history_length_s * estimated_frame_rate))
        self.intra_history = np.zeros(self.history_length_n)
        self.inter_history = np.zeros(self.history_length_n)

        self.num_sectors = min(processor_config.num_zones, sensor_config.num_points)
        self.sector_size = max(1, -(-sensor_config.num_points // self.num_sectors))

        self.sector_offset = (self.num_sectors * self.sector_size - sensor_config.num_points) // 2
        self.zone_limits = zone_limits

        self.setup_is_done = False

    def setup(self, win):
        win.setWindowTitle("Acconeer smart presence example")

        self.intra_limit_lines = []
        self.inter_limit_lines = []

        # Intra presence history plot

        self.intra_hist_plot = win.addPlot(
            row=0,
            col=0,
            title="Intra presence history (fast motions)",
        )
        self.intra_hist_plot.setMenuEnabled(False)
        self.intra_hist_plot.setMouseEnabled(x=False, y=False)
        self.intra_hist_plot.hideButtons()
        self.intra_hist_plot.showGrid(x=True, y=True)
        self.intra_hist_plot.setLabel("bottom", "Time (s)")
        self.intra_hist_plot.setLabel("left", "Score")
        self.intra_hist_plot.setXRange(-self.history_length_s, 0)
        self.intra_history_smooth_max = et.utils.SmoothMax(self.estimated_frame_rate)
        self.intra_hist_plot.setYRange(0, 10)
        if not self.detector_config.intra_enable:
            intra_color = et.utils.color_cycler(1)
            intra_color = f"{intra_color}50"
            intra_dashed_pen = pg.mkPen(intra_color, width=2.5, style=QtCore.Qt.DashLine)
            intra_pen = pg.mkPen(intra_color, width=2)
        else:
            intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
            intra_pen = et.utils.pg_pen_cycler(1)

        self.intra_hist_curve = self.intra_hist_plot.plot(pen=intra_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=intra_dashed_pen)
        self.intra_hist_plot.addItem(limit_line)
        self.intra_limit_lines.append(limit_line)

        for line in self.intra_limit_lines:
            line.setPos(self.detector_config.intra_detection_threshold)

        # Inter presence history plot

        self.inter_hist_plot = win.addPlot(
            row=0,
            col=1,
            title="Inter presence history (slow motions)",
        )
        self.inter_hist_plot.setMenuEnabled(False)
        self.inter_hist_plot.setMouseEnabled(x=False, y=False)
        self.inter_hist_plot.hideButtons()
        self.inter_hist_plot.showGrid(x=True, y=True)
        self.inter_hist_plot.setLabel("bottom", "Time (s)")
        self.inter_hist_plot.setLabel("left", "Score")
        self.inter_hist_plot.setXRange(-self.history_length_s, 0)
        self.inter_history_smooth_max = et.utils.SmoothMax(self.estimated_frame_rate)
        self.inter_hist_plot.setYRange(0, 10)
        if not self.detector_config.inter_enable:
            inter_color = et.utils.color_cycler(0)
            inter_color = f"{inter_color}50"
            inter_dashed_pen = pg.mkPen(inter_color, width=2.5, style=QtCore.Qt.DashLine)
            inter_pen = pg.mkPen(inter_color, width=2)
        else:
            inter_pen = et.utils.pg_pen_cycler(0)
            inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

        self.inter_hist_curve = self.inter_hist_plot.plot(pen=inter_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=inter_dashed_pen)
        self.inter_hist_plot.addItem(limit_line)
        self.inter_limit_lines.append(limit_line)

        for line in self.inter_limit_lines:
            line.setPos(self.detector_config.inter_detection_threshold)

        # Sector plot

        self.sector_plot = pg.PlotItem(
            title="Detection zone<br>Detection type: fast (orange), slow (blue), both (green)"
        )
        self.sector_plot.setAspectLocked()
        self.sector_plot.hideAxis("left")
        self.sector_plot.hideAxis("bottom")
        self.sectors = []
        self.limit_text = []

        self.range_html = (
            '<div style="text-align: center">'
            '<span style="color: #000000;font-size:12pt;">'
            "{}</span></div>"
        )

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(self.num_sectors) + 1):
            sector = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            self.sector_plot.addItem(sector)
            self.sectors.append(sector)

            limit = pg.TextItem(html=self.range_html, anchor=(0.5, 0.5), angle=25)
            x = r * np.cos(np.radians(span_deg))
            y = r * np.sin(np.radians(span_deg))
            limit.setPos(x, y + 0.25)
            self.sector_plot.addItem(limit)
            self.limit_text.append(limit)

        self.sectors.reverse()

        start_limit_text = pg.TextItem(html=self.range_html, anchor=(0.5, 0.5), angle=25)
        range_html = self.range_html.format(f"{self.detector_config.start_m}")
        start_limit_text.setHtml(range_html)
        start_limit_text.setPos(0, 0.25)
        self.sector_plot.addItem(start_limit_text)

        unit_text = pg.TextItem(html=self.range_html, anchor=(0.5, 0.5))
        unit_html = self.range_html.format("[m]")
        unit_text.setHtml(unit_html)
        unit_text.setPos(
            self.num_sectors + 0.5, (self.num_sectors + 1) * np.sin(np.radians(span_deg))
        )
        self.sector_plot.addItem(unit_text)

        for (text_item, limit) in zip(self.limit_text, np.flip(self.zone_limits)):
            range_html = self.range_html.format(np.around(limit, 1))
            text_item.setHtml(range_html)

        sublayout = win.addLayout(row=1, col=0, colspan=2)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.sector_plot, row=0, col=0)

        self.setup_is_done = True

    def update(self, data):
        processor_result = data["processor_result"]
        detector_result = data["detector_result"]

        # Intra presence

        move_hist_xs = np.linspace(-self.history_length_s, 0, self.history_length_n)

        self.intra_history = np.roll(self.intra_history, -1)
        self.intra_history[-1] = detector_result.intra_presence_score

        m_hist = max(
            float(np.max(self.intra_history)),
            self.detector_config.intra_detection_threshold * 1.05,
        )
        m_hist = self.intra_history_smooth_max.update(m_hist)

        self.intra_hist_plot.setYRange(0, m_hist)
        self.intra_hist_curve.setData(move_hist_xs, self.intra_history)

        # Inter presence

        self.inter_history = np.roll(self.inter_history, -1)
        self.inter_history[-1] = detector_result.inter_presence_score

        m_hist = max(
            float(np.max(self.inter_history)),
            self.detector_config.inter_detection_threshold * 1.05,
        )
        m_hist = self.inter_history_smooth_max.update(m_hist)

        self.inter_hist_plot.setYRange(0, m_hist)
        self.inter_hist_curve.setData(move_hist_xs, self.inter_history)

        # Sector

        brush = et.utils.pg_brush_cycler(7)
        for sector in self.sectors:
            sector.setBrush(brush)

        if detector_result.presence_detected:
            assert processor_result.max_presence_zone is not None
            if processor_result.max_presence_zone == processor_result.max_intra_zone:
                self.sectors[processor_result.max_presence_zone].setBrush(
                    et.utils.pg_brush_cycler(1)
                )
            else:
                self.sectors[processor_result.max_presence_zone].setBrush(
                    et.utils.pg_brush_cycler(0)
                )


if __name__ == "__main__":
    main()
